import os
import re

import pytest
from dbt.tests.adapter.ephemeral.test_ephemeral import (
    BaseEphemeral,
    BaseEphemeralMulti,
    ephemeral_errors__base__base_copy_sql,
    ephemeral_errors__base__base_sql,
    ephemeral_errors__dependent_sql,
    models_n__ephemeral_level_two_sql,
    models_n__ephemeral_sql,
    models_n__root_view_sql,
    models_n__source_table_sql,
)
from dbt.tests.util import check_relations_equal, run_dbt


class TestEphemeralMulti(BaseEphemeralMulti):
    def test_ephemeral_multi(self, project):
        db = project.database

        run_dbt(["seed"])
        results = run_dbt(["run"])
        assert len(results) == 3

        check_relations_equal(project.adapter, ["seed", "dependent"])
        check_relations_equal(project.adapter, ["seed", "double_dependent"])
        check_relations_equal(project.adapter, ["seed", "super_dependent"])
        assert os.path.exists("./target/run/test/models/double_dependent.sql")
        with open("./target/run/test/models/double_dependent.sql", "r") as fp:
            sql_file = fp.read()

        sql_file = re.sub(r"\d+", "", sql_file)
        expected_sql = (
            f'create view "{db}"."test_test_ephemeral"."double_dependent__dbt_tmp" as ('
            "with __dbt__cte__base as ("
            "select * from test_test_ephemeral.seed"
            "),  __dbt__cte__base_copy as ("
            "select * from __dbt__cte__base"
            ")-- base_copy just pulls from base. Make sure the listed"
            "-- graph of CTEs all share the same dbt_cte__base cte"
            "select * from __dbt__cte__base where gender = 'Male'"
            "union all"
            "select * from __dbt__cte__base_copy where gender = 'Female'"
            ");"
        )
        sql_file = "".join(sql_file.split())
        expected_sql = "".join(expected_sql.split())
        assert sql_file == expected_sql


class TestEphemeralNested(BaseEphemeral):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "ephemeral_level_two.sql": models_n__ephemeral_level_two_sql,
            "root_view.sql": models_n__root_view_sql,
            "ephemeral.sql": models_n__ephemeral_sql,
            "source_table.sql": models_n__source_table_sql,
        }

    def test_ephemeral_nested(self, project):
        db = project.database

        results = run_dbt(["run"])
        assert len(results) == 2
        assert os.path.exists("./target/run/test/models/root_view.sql")
        with open("./target/run/test/models/root_view.sql", "r") as fp:
            sql_file = fp.read()

        sql_file = re.sub(r"\d+", "", sql_file)
        expected_sql = (
            f'create view "{db}"."test_test_ephemeral"."root_view__dbt_tmp" as ('
            "with __dbt__cte__ephemeral_level_two as ("
            f'select * from "{db}"."test_test_ephemeral"."source_table"'
            "),  __dbt__cte__ephemeral as ("
            "select * from __dbt__cte__ephemeral_level_two"
            ")select * from __dbt__cte__ephemeral"
            ");"
        )

        sql_file = "".join(sql_file.split())
        expected_sql = "".join(expected_sql.split())
        assert sql_file == expected_sql


class TestEphemeralErrorHandling(BaseEphemeral):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "dependent.sql": ephemeral_errors__dependent_sql,
            "base": {
                "base.sql": ephemeral_errors__base__base_sql,
                "base_copy.sql": ephemeral_errors__base__base_copy_sql,
            },
        }

    def test_ephemeral_error_handling(self, project):
        results = run_dbt(["run"], expect_pass=False)
        assert len(results) == 1
        assert results[0].status == "skipped"
        assert "Compilation Error" in results[0].message



import os
import pytest
from dbt.tests.util import run_dbt, relation_from_name
from dbt.adapters.duckdb import DuckDBConnectionManager

upstream_model_sql = """
select range from range(3)
"""

upstream_partition_by_model = """
{{ config(materialized='external', options={"partition_by": "a"}) }}
select range as a, 'foo' as b from range(5)
"""

downstream_model_sql = """
select range * 2 from {{ ref('upstream_model') }}
"""

other_downstream_model_sql = """
select range * 5 from {{ ref('upstream_model') }}
"""

downstream_of_partition_model = """
select a from {{ ref('upstream_partition_by_model') }}
"""


# class must begin with 'Test'
class TestRematerializeDownstreamExternalModel:
    """
    External models should load in dependencies when they exist.

    We test that after materializing upstream and downstream models, we can
    materialize the downstream model by itself, even if we are using an
    in-memory database.
    """

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target, tmp_path_factory):
        extroot = str(tmp_path_factory.getbasetemp() / "rematerialize")
        os.mkdir(extroot)
        dbt_profile_target["external_root"] = extroot
        return dbt_profile_target
    
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "base",
            "models": {"+materialized": "external"},
            "on-run-start": ["{{ register_upstream_external_models() }}"],
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_model.sql": upstream_model_sql,
            "upstream_partition_by_model.sql": upstream_partition_by_model,
            "downstream_model.sql": downstream_model_sql,
            "other_downstream_model.sql": other_downstream_model_sql,
            "downstream_of_partition_model.sql": downstream_of_partition_model,
        }

    def test_run(self, project):
        run_dbt(["run"])

        # Force close the :memory: connection
        DuckDBConnectionManager.close_all_connections()
        run_dbt(
            [
                "run",
                "--select",
                "downstream_model other_downstream_model downstream_of_partition_model",
            ]
        )

        # really makes sure we have created the downstream model
        relation = relation_from_name(project.adapter, "downstream_of_partition_model")
        result = project.run_sql(f"select count(*) as num_rows from {relation}", fetch="one")
        assert result[0] == 5



from urllib.parse import urlparse
import pytest
from dbt.tests.util import (
    run_dbt,
)

random_logs_sql = """
{{ config(materialized='table', meta=dict(temp_schema_name='dbt_temp_test')) }}

select 
  uuid()::varchar as log_id,
 '2023-10-01'::timestamp + interval 1 minute * (random() * 20000)::int as dt ,
(random() * 4)::int64 as user_id
from generate_series(1, 10000) g(x)
"""

summary_of_logs_sql = """
{{
    config(
        materialized='incremental',
        meta=dict(temp_schema_name='dbt_temp_test'),
    )
}}

select dt::date as dt, user_id, count(1) as c
from {{ ref('random_logs_test') }}


{% if is_incremental() %}

  -- this filter will only be applied on an incremental run
  -- (uses > to include records whose timestamp occurred since the last run of this model)
  where dt > '2023-10-08'::timestamp

{% endif %}
group by all
"""

python_pyarrow_table_model = """
import pyarrow as pa

def model(dbt, con):
    return pa.Table.from_pydict({"a": [1,2,3]})
"""

@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMDPluginAttach:
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        md_config = {"token": dbt_profile_target.get("token")}
        plugins = [{"module": "motherduck", "config": md_config}]
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": ":memory:",
                        "plugins": plugins,
                        "attach": [
                            {
                                "path": dbt_profile_target.get("path", ":memory:"),
                                "type": "motherduck"
                            }
                        ]
                    }
                },
                "target": "dev",
            }
        }
    
    @pytest.fixture(scope="class")
    def models(self, md_sql):
        return {
            "md_table.sql": md_sql,
            "random_logs_test.sql": random_logs_sql,
            "summary_of_logs_test.sql": summary_of_logs_sql,
            "python_pyarrow_table_model.py": python_pyarrow_table_model,
        }
    
    @pytest.fixture(scope="class")
    def database_name(self, dbt_profile_target):
        return urlparse(dbt_profile_target["path"]).path
    
    @pytest.fixture(scope="class")
    def md_sql(self, database_name):
        # Reads from a MD database in my test account in the cloud
        return f"""
            select * FROM {database_name}.main.plugin_table
        """
    
    @pytest.fixture(autouse=True)
    def run_dbt_scope(self, project, database_name):
        project.run_sql(f"CREATE DATABASE IF NOT EXISTS {database_name}")
        project.run_sql(f"CREATE OR REPLACE TABLE {database_name}.plugin_table (i integer, j string)")
        project.run_sql(f"INSERT INTO {database_name}.plugin_table (i, j) VALUES (1, 'foo')")
        yield
        project.run_sql("DROP VIEW md_table")
        project.run_sql("DROP TABLE random_logs_test")
        project.run_sql("DROP TABLE summary_of_logs_test")
        project.run_sql(f"DROP TABLE {database_name}.plugin_table")
        project.run_sql("DROP TABLE python_pyarrow_table_model")

    def test_motherduck(self, project):
        run_dbt(expect_pass=True)


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMDPluginAttachWithSettings(TestMDPluginAttach):
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        md_setting = {"motherduck_token": dbt_profile_target.get("token")}
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": ":memory:",
                        "attach": [
                            {
                                "path": dbt_profile_target.get("path", ":memory:"),
                                "type": "motherduck"
                            }
                        ],
                        "settings": md_setting
                    }
                },
                "target": "dev",
            }
        }


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMDPluginAttachWithTokenInPath(TestMDPluginAttach):
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        token = dbt_profile_target.get("token")
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": ":memory:",
                        "attach": [
                            {
                                "path": dbt_profile_target.get("path", ":memory:") + f"?motherduck_token={token}&user=1",
                                "type": "motherduck"
                            }
                        ]
                    }
                },
                "target": "dev",
            }
        }



import unittest
from argparse import Namespace
from unittest import mock

from dbt.flags import set_from_args
from dbt.adapters.duckdb import DuckDBAdapter
from tests.unit.utils import config_from_parts_or_dicts

class TestExternalUtils(unittest.TestCase):
    def setUp(self):
        set_from_args(Namespace(STRICT_MODE=True), {})

        profile_cfg = {
            "outputs": {
                "test": {
                    "type": "duckdb",
                    "path": ":memory:",
                }
            },
            "target": "test",
        }

        project_cfg = {
            "name": "X",
            "version": "0.1",
            "profile": "test",
            "project-root": "/tmp/dbt/does-not-exist",
            "config-version": 2,
        }

        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg, cli_vars={})
        self._adapter = None

    @property
    def adapter(self):
        self.mock_mp_context = mock.MagicMock()
        if self._adapter is None:
            self._adapter = DuckDBAdapter(self.config, self.mock_mp_context)
        return self._adapter

    def test_external_write_options(self):
        data = [
            ("/tmp/test.csv", {}, "format csv, header 1"),
            ("./foo.parquet", {"codec": "zstd"}, "codec zstd, format parquet"),
            ("bar", {"delimiter": "|", "header": "0"}, "delimiter '|', header 0, format csv"),
            ("a.parquet", {"partition_by": "ds"}, "partition_by ds, format parquet"),
            ("b.csv", {"partition_by": "ds,category"}, "partition_by (ds,category), format csv, header 1"),
            ("/path/to/c.csv", {"null": "\\N"}, "null '\\N', format csv, header 1")
        ]

        for (loc, opts, expected) in data:
            assert expected == self.adapter.external_write_options(loc, opts)
    

    def test_external_read_location(self):
        data = [
            ("bar", {"format": "csv", "delimiter": "|", "header": "0"}, "bar"),
            ("/tmp/a", {"partition_by": "ds", "format": "parquet"}, "/tmp/a/*/*.parquet"),
            ("b", {"partition_by": "ds,category"}, "b/*/*/*.parquet"),
        ]
        for (loc, opts, expected) in data:
            assert expected == self.adapter.external_read_location(loc, opts)


import unittest
from argparse import Namespace
from unittest import mock

from dbt.flags import set_from_args
from dbt.adapters.duckdb import DuckDBAdapter
from dbt.adapters.duckdb.connections import DuckDBConnectionManager
from tests.unit.utils import config_from_parts_or_dicts, mock_connection


class TestDuckDBAdapter(unittest.TestCase):
    def setUp(self):
        set_from_args(Namespace(STRICT_MODE=True), {})

        profile_cfg = {
            "outputs": {
                "test": {
                    "type": "duckdb",
                    "path": ":memory:",
                }
            },
            "target": "test",
        }

        project_cfg = {
            "name": "X",
            "version": "0.1",
            "profile": "test",
            "project-root": "/tmp/dbt/does-not-exist",
            "quoting": {
                "identifier": False,
                "schema": True,
            },
            "config-version": 2,
        }

        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg, cli_vars={})
        self._adapter = None

    @property
    def adapter(self):
        self.mock_mp_context = mock.MagicMock()
        if self._adapter is None:
            self._adapter = DuckDBAdapter(self.config, self.mock_mp_context)
        return self._adapter

    @mock.patch("dbt.adapters.duckdb.environments.duckdb")
    def test_acquire_connection(self, connector):
        connector.__version__ = "0.1.0"  # dummy placeholder for semver checks
        DuckDBConnectionManager.close_all_connections()
        connection = self.adapter.acquire_connection("dummy")

        connector.connect.assert_not_called()
        connection.handle
        self.assertEqual(connection.state, "open")
        self.assertNotEqual(connection.handle, None)
        connector.connect.assert_called_once()

    def test_cancel_open_connections_empty(self):
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_main(self):
        key = self.adapter.connections.get_thread_identifier()
        self.adapter.connections.thread_connections[key] = mock_connection("main")
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)


class TestDuckDBAdapterWithSecrets(unittest.TestCase):
    def setUp(self):
        set_from_args(Namespace(STRICT_MODE=True), {})

        profile_cfg = {
            "outputs": {
                "test": {
                    "type": "duckdb",
                    "path": ":memory:",
                    "secrets": [
                        {
                            "type": "s3",
                            "key_id": "abc",
                            "secret": "xyz",
                            "region": "us-west-2"
                        }
                    ]
                }
            },
            "target": "test",
        }

        project_cfg = {
            "name": "X",
            "version": "0.1",
            "profile": "test",
            "project-root": "/tmp/dbt/does-not-exist",
            "quoting": {
                "identifier": False,
                "schema": True,
            },
            "config-version": 2,
        }

        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg, cli_vars={})
        self._adapter = None

    @property
    def adapter(self):
        self.mock_mp_context = mock.MagicMock()
        if self._adapter is None:
            self._adapter = DuckDBAdapter(self.config, self.mock_mp_context)
        return self._adapter

    @mock.patch("dbt.adapters.duckdb.environments.duckdb")
    def test_create_secret(self, connector):
        connector.__version__ = "0.1.0"  # dummy placeholder for semver checks
        DuckDBConnectionManager.close_all_connections()
        connection = self.adapter.acquire_connection("dummy")
        assert connection.handle
        connection._handle._env.conn.execute.assert_called_with(
"""CREATE OR REPLACE SECRET _dbt_secret_1 (
    type s3,
    key_id 'abc',
    secret 'xyz',
    region 'us-west-2'
)""")



import os
from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Union
from urllib.parse import urlparse

from dbt_common.dataclass_schema import dbtClassMixin
from dbt_common.exceptions import DbtRuntimeError

from dbt.adapters.contracts.connection import Credentials
from dbt.adapters.duckdb.secrets import DEFAULT_SECRET_PREFIX
from dbt.adapters.duckdb.secrets import Secret


@dataclass
class Attachment(dbtClassMixin):
    # The path to the database to be attached (may be a URL)
    path: str

    # The type of the attached database (defaults to duckdb, but may be supported by an extension)
    type: Optional[str] = None

    # An optional alias for the attached database
    alias: Optional[str] = None

    # Whether the attached database is read-only or read/write
    read_only: bool = False

    def to_sql(self) -> str:
        # remove query parameters (not supported in ATTACH)
        parsed = urlparse(self.path)
        path = self.path.replace(f"?{parsed.query}", "")
        base = f"ATTACH '{path}'"
        if self.alias:
            base += f" AS {self.alias}"
        options = []
        if self.type:
            options.append(f"TYPE {self.type}")
        if self.read_only:
            options.append("READ_ONLY")
        if options:
            joined = ", ".join(options)
            base += f" ({joined})"
        return base


@dataclass
class PluginConfig(dbtClassMixin):
    module: str

    alias: Optional[str] = None

    # A plugin-specific set of configuration options
    config: Optional[Dict[str, Any]] = None


@dataclass
class Remote(dbtClassMixin):
    host: str
    port: int
    user: str
    password: Optional[str] = None


@dataclass
class Retries(dbtClassMixin):
    # The number of times to attempt the initial duckdb.connect call
    # (to wait for another process to free the lock on the DB file)
    connect_attempts: int = 1

    # The number of times to attempt to execute a DuckDB query that throws
    # one of the retryable exceptions
    query_attempts: Optional[int] = None

    # The list of exceptions that we are willing to retry on
    retryable_exceptions: List[str] = field(default_factory=lambda: ["IOException"])


@dataclass
class Extension(dbtClassMixin):
    name: str
    repo: str


@dataclass
class DuckDBCredentials(Credentials):
    database: str = "main"
    schema: str = "main"
    path: str = ":memory:"

    # Any connection-time configuration information that we need to pass
    # to DuckDB (e.g., if we need to enable using unsigned extensions)
    config_options: Optional[Dict[str, Any]] = None

    # any DuckDB extensions we want to install and load (httpfs, parquet, etc.)
    extensions: Optional[List[Union[str, Dict[str, str]]]] = None

    # any additional pragmas we want to configure on our DuckDB connections;
    # a list of the built-in pragmas can be found here:
    # https://duckdb.org/docs/sql/configuration
    # (and extensions may add their own pragmas as well)
    settings: Optional[Dict[str, Any]] = None

    # secrets for connecting to cloud services AWS S3, Azure, Cloudfare R2,
    # Google Cloud and Huggingface.
    secrets: Optional[List[Dict[str, Any]]] = None

    # the root path to use for any external materializations that are specified
    # in this dbt project; defaults to "." (the current working directory)
    external_root: str = "."

    # identify whether to use the default credential provider chain for AWS/GCloud
    # instead of statically defined environment variables
    use_credential_provider: Optional[str] = None

    # A list of additional databases that should be attached to the running
    # DuckDB instance to make them available for use in models; see the
    # schema for the Attachment dataclass above for what fields it can contain
    attach: Optional[List[Attachment]] = None

    # A list of filesystems to attach to the DuckDB database via the fsspec
    # interface; see https://duckdb.org/docs/guides/python/filesystems.html
    #
    # Each dictionary entry must have a "fs" entry to indicate which
    # fsspec implementation should be loaded, and then an arbitrary additional
    # number of key-value pairs that will be passed as arguments to the fsspec
    # registry method.
    filesystems: Optional[List[Dict[str, Any]]] = None

    # Used to configure remote environments/connections
    remote: Optional[Remote] = None

    # A list of dbt-duckdb plugins that can be used to customize the
    # behavior of loading source data and/or storing the relations that are
    # created by SQL or Python models; see the plugins module for more details.
    plugins: Optional[List[PluginConfig]] = None

    # Whether to disable transactions when executing SQL statements; this
    # is useful when we would like the resulting DuckDB database file to
    # be as small as possible.
    disable_transactions: bool = False

    # Whether to keep the DuckDB connection open between invocations of dbt
    keep_open: bool = True

    # A list of paths to Python modules that should be loaded into the
    # running Python environment when dbt is invoked; this is useful for
    # loading custom dbt-duckdb plugins or locally defined modules that
    # provide helper functions for dbt Python models.
    module_paths: Optional[List[str]] = None

    # An optional strategy for allowing retries when certain types of
    # exceptions occur on a model run (e.g., IOExceptions that were caused
    # by networking issues)
    retries: Optional[Retries] = None

    def __post_init__(self):
        self.settings = self.settings or {}
        self.secrets = self.secrets or []
        self._secrets = []

        # Add MotherDuck plugin if the path is a MotherDuck database
        # and plugin was not specified in profile.yml
        if self.is_motherduck:
            if self.plugins is None:
                self.plugins = []
            if "motherduck" not in [plugin.module for plugin in self.plugins]:
                self.plugins.append(PluginConfig(module="motherduck"))

        # For backward compatibility, to be deprecated in the future
        if self.use_credential_provider:
            if self.use_credential_provider == "aws":
                self.secrets.append({"type": "s3", "provider": "credential_chain"})
            else:
                raise ValueError(
                    "Unsupported value for use_credential_provider: "
                    + self.use_credential_provider
                )

        if self.secrets:
            self._secrets = [
                Secret.create(
                    secret_type=secret.pop("type"),
                    name=secret.pop("name", f"{DEFAULT_SECRET_PREFIX}{num + 1}"),
                    **secret,
                )
                for num, secret in enumerate(self.secrets)
            ]

    def secrets_sql(self) -> List[str]:
        return [secret.to_sql() for secret in self._secrets]

    @property
    def motherduck_attach(self):
        # Check if any MotherDuck paths are attached
        attach = []
        for attached_db in self.attach or []:
            parsed = urlparse(attached_db.path)
            if self._is_motherduck(parsed.scheme):
                attach.append(attached_db)
        return attach

    @property
    def is_motherduck_attach(self):
        return len(self.motherduck_attach) > 0

    @property
    def is_motherduck(self):
        parsed = urlparse(self.path)
        return self._is_motherduck(parsed.scheme) or self.is_motherduck_attach

    @staticmethod
    def _is_motherduck(scheme: str) -> bool:
        return scheme in {"md", "motherduck"}

    @classmethod
    def __pre_deserialize__(cls, data: Dict[Any, Any]) -> Dict[Any, Any]:
        data = super().__pre_deserialize__(data)
        path = data.get("path")
        path_db = None
        if path is None or path == ":memory:":
            path_db = "memory"
        else:
            parsed = urlparse(path)
            base_file = os.path.basename(parsed.path)
            path_db = os.path.splitext(base_file)[0]
            # For MotherDuck, turn on disable_transactions unless
            # it's explicitly set already by the user
            if cls._is_motherduck(parsed.scheme):
                if "disable_transactions" not in data:
                    data["disable_transactions"] = True
                if path_db == "":
                    path_db = "my_db"

        if path_db and "database" not in data:
            data["database"] = path_db
        elif path_db and data["database"] != path_db:
            if not data.get("remote"):
                raise DbtRuntimeError(
                    "Inconsistency detected between 'path' and 'database' fields in profile; "
                    f"the 'database' property must be set to '{path_db}' to match the 'path'"
                )
        elif not path_db:
            raise DbtRuntimeError(
                "Unable to determine target database name from 'path' field in profile"
            )
        return data

    @property
    def unique_field(self) -> str:
        """
        This property returns a unique field for the database connection.
        If the connection is remote, it returns the host and port as a string.
        If the connection is local, it returns the path and external root as a string.
        """
        if self.remote:
            return self.remote.host + str(self.remote.port)
        else:
            return self.path + self.external_root

    @property
    def type(self):
        return "duckdb"

    def _connection_keys(self):
        return (
            "database",
            "schema",
            "path",
            "config_options",
            "extensions",
            "settings",
            "external_root",
            "use_credential_provider",
            "attach",
            "filesystems",
            "remote",
            "plugins",
            "disable_transactions",
        )



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

from duckdb import DuckDBPyConnection

from . import BasePlugin
from dbt.adapters.events.logging import AdapterLogger

PG_EXT = "postgres"


class Plugin(BasePlugin):
    logger = AdapterLogger("DuckDB_PostgresPlugin")

    def __init__(self, name: str, plugin_config: Dict[str, Any]):
        """
        Initialize the Plugin with a name and configuration.
        """
        super().__init__(name, plugin_config)
        self.logger.debug(
            "Plugin __init__ called with name: %s and config: %s", name, plugin_config
        )
        self.initialize(plugin_config)

    def initialize(self, config: Dict[str, Any]):
        """
        Initialize the plugin with the provided configuration.
        """
        self.logger.debug("Initializing PostgreSQL plugin with config: %s", config)

        self._dsn: str = config["dsn"]
        if not self._dsn:
            self.logger.error(
                "Initialization failed: 'dsn' is a required argument for the postgres plugin!"
            )
            raise ValueError("'dsn' is a required argument for the postgres plugin!")

        self._pg_schema: Optional[str] = config.get("pg_schema")  # Can be None
        self._duckdb_alias: str = config.get("duckdb_alias", "postgres_db")
        self._read_only: bool = config.get("read_only", False)
        self._secret: Optional[str] = config.get("secret")
        self._attach_options: Dict[str, Any] = config.get(
            "attach_options", {}
        )  # Additional ATTACH options
        self._settings: Dict[str, Any] = config.get(
            "settings", {}
        )  # Extension settings via SET commands

        self.logger.info(
            "PostgreSQL plugin initialized with dsn='%s', pg_schema='%s', "
            "duckdb_alias='%s', read_only=%s, secret='%s'",
            self._dsn,
            self._pg_schema,
            self._duckdb_alias,
            self._read_only,
            self._secret,
        )

    def configure_connection(self, conn: DuckDBPyConnection):
        """
        Configure the DuckDB connection to attach the PostgreSQL database.
        """
        self.logger.debug("Configuring DuckDB connection for PostgreSQL plugin.")

        conn.install_extension(PG_EXT)
        conn.load_extension(PG_EXT)
        self.logger.info("PostgreSQL extension installed and loaded.")

        # Set any extension settings provided
        self._set_extension_settings(conn)

        # Build and execute the ATTACH command
        attach_stmt = self._build_attach_statement()
        self.logger.debug("Executing ATTACH statement: %s", attach_stmt)
        try:
            conn.execute(attach_stmt)
            self.logger.info("Successfully attached PostgreSQL database with DSN: %s", self._dsn)
        except Exception as e:
            self.logger.error("Failed to attach PostgreSQL database: %s", e)
            raise

    def _set_extension_settings(self, conn: DuckDBPyConnection):
        """
        Set extension settings via SET commands.
        """
        for setting, value in self._settings.items():
            # Quote string values
            if isinstance(value, str):
                value = f"'{value}'"
            elif isinstance(value, bool):
                value = "true" if value else "false"
            set_stmt = f"SET {setting} = {value};"
            self.logger.debug("Setting extension option: %s", set_stmt)
            try:
                conn.execute(set_stmt)
            except Exception as e:
                self.logger.error("Failed to set option %s: %s", setting, e)
                raise

    def _build_attach_statement(self) -> str:
        """
        Build the ATTACH statement for connecting to the PostgreSQL database.
        """
        attach_options: List[Tuple[str, Optional[str]]] = [("TYPE", "POSTGRES")]

        if self._pg_schema:
            attach_options.append(("SCHEMA", f"'{self._pg_schema}'"))

        if self._secret:
            attach_options.append(("SECRET", f"'{self._secret}'"))

        # Additional attach options
        for k, v in self._attach_options.items():
            if isinstance(v, bool):
                v = "true" if v else "false"
            elif isinstance(v, str):
                v = f"'{v}'"
            attach_options.append((k.upper(), v))

        if self._read_only:
            attach_options.append(("READ_ONLY", None))  # No value assigned

        # Convert options to string
        attach_options_str = ", ".join(
            f"{k} {v}" if v is not None else k for k, v in attach_options
        )

        attach_stmt = f"ATTACH '{self._dsn}' AS {self._duckdb_alias} ({attach_options_str});"
        return attach_stmt



import os
import resource
import subprocess
import time
from importlib import metadata

import duckdb
import pytest

# Increase the number of open files allowed
# Hack for https://github.com/dbt-labs/dbt-core/issues/7316
soft_limit, hard_limit = resource.getrlimit(resource.RLIMIT_NOFILE)
resource.setrlimit(resource.RLIMIT_NOFILE, (hard_limit, hard_limit))

# Import the standard functional fixtures as a plugin
# Note: fixtures with session scope need to be local
pytest_plugins = ["dbt.tests.fixtures.project"]

MOTHERDUCK_TOKEN = "MOTHERDUCK_TOKEN"
TEST_MOTHERDUCK_TOKEN = "TEST_MOTHERDUCK_TOKEN"


def pytest_addoption(parser):
    parser.addoption("--profile", action="store", default="memory", type=str)


def pytest_report_header() -> list[str]:
    """Return a list of strings to be displayed in the header of the report."""
    return [
        f"duckdb: {metadata.version('duckdb')}",
        f"dbt-core: {metadata.version('dbt-core')}",
    ]


@pytest.fixture(scope="session")
def profile_type(request):
    return request.config.getoption("--profile")


@pytest.fixture(scope="session")
def bv_server_process(profile_type):
    if profile_type == "buenavista":
        server_process = subprocess.Popen(["python3", "-m", "tests.bv_test_server"])

        # Wait for the server to be ready
        time.sleep(5)

        # Pass the server process to the tests
        yield server_process

        # Teardown: Stop the server process after tests are done
        server_process.terminate()
        server_process.wait()
    else:
        yield None


# The profile dictionary, used to write out profiles.yml
# dbt will supply a unique schema per test, so we do not specify 'schema' here
@pytest.fixture(scope="session")
def dbt_profile_target(profile_type, bv_server_process, tmpdir_factory):
    profile = {"type": "duckdb", "threads": 4}

    if profile_type == "buenavista":
        profile["database"] = "memory"
        profile["remote"] = {
            "host": "127.0.0.1",
            "port": 5433,
            "user": "test",
        }
    elif profile_type == "file":
        profile["path"] = str(tmpdir_factory.mktemp("dbs") / "tmp.db")
    elif profile_type == "md":
        # Test against MotherDuck
        if MOTHERDUCK_TOKEN not in os.environ and MOTHERDUCK_TOKEN.lower() not in os.environ:
            if TEST_MOTHERDUCK_TOKEN not in os.environ:
                raise ValueError(
                    f"Please set the {MOTHERDUCK_TOKEN} or {TEST_MOTHERDUCK_TOKEN} \
                        environment variable to run tests against MotherDuck"
                )
            profile["token"] = os.environ.get(TEST_MOTHERDUCK_TOKEN)
        else:
            profile["token"] = os.environ.get(MOTHERDUCK_TOKEN, os.environ.get(MOTHERDUCK_TOKEN.lower()))
        profile["disable_transactions"] = True
        profile["path"] = "md:test"
    elif profile_type in ["memory", "nightly"]:
        pass  # use the default path-less profile
    else:
        raise ValueError(f"Invalid profile type '{profile_type}'")

    return profile


@pytest.fixture(autouse=True, scope="class")
def skip_by_profile_type(profile_type, request):
    if request.node.get_closest_marker("skip_profile"):
        for skip_profile_type in request.node.get_closest_marker("skip_profile").args:
            if skip_profile_type == profile_type:
                pytest.skip(f"skipped on '{profile_type}' profile")


@pytest.fixture(scope="session")
def test_data_path():
    test_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(test_dir, "data")


def pytest_collection_modifyitems(config, items):
    # Skip the S3 tests if the secrets are not available
    if not (
        os.getenv("S3_MD_ORG_KEY") and os.getenv("S3_MD_ORG_REGION") and os.getenv("S3_MD_ORG_SECRET")
    ):
        skip_s3 = pytest.mark.skip(reason="need S3 credentials to run this test")
        for item in items:
            if "with_s3_creds" in item.keywords:
                item.add_marker(skip_s3)



import os
import tempfile

import duckdb
import pytest

from dbt.tests.util import run_dbt

sources_schema_yml = """
version: 2
sources:
  - name: attached_source
    database: attach_test
    schema: analytics
    tables:
      - name: attached_table
        description: "An attached table"
        columns:
          - name: id
            description: "An id"
            tests:
              - unique
              - not_null
"""

models_source_model_sql = """
    select * from {{ source('attached_source', 'attached_table') }}
"""

models_target_model_sql = """
    {{ config(materialized='table', database='attach_test') }}
    SELECT * FROM {{ ref('source_model') }}
"""


@pytest.mark.skip_profile("memory", "buenavista", "md")
class TestAttachedDatabase:
    @pytest.fixture(scope="class")
    def attach_test_db(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            path = os.path.join(temp_dir, "attach_test.duckdb")
            db = duckdb.connect(path)
            db.execute("CREATE SCHEMA analytics")
            db.execute("CREATE TABLE analytics.attached_table AS SELECT 1 as id")
            db.close()
            yield path

    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target, attach_test_db):
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target.get("path", ":memory:"),
                        "attach": [{"path": attach_test_db}],
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": sources_schema_yml,
            "source_model.sql": models_source_model_sql,
            "target_model.sql": models_target_model_sql,
        }

    def test_attached_databases(self, project, attach_test_db):
        results = run_dbt()
        assert len(results) == 2

        test_results = run_dbt(["test"])
        assert len(test_results) == 2

        # check that the model is created in the attached db
        db = duckdb.connect(attach_test_db)
        ret = db.execute("SELECT * FROM target_model").fetchall()
        assert ret[0][0] == 1
        db.close()

        # check that everything works on a re-run of dbt
        rerun_results = run_dbt()
        assert len(rerun_results) == 2



from dbt.tests.adapter.empty.test_empty import BaseTestEmpty, BaseTestEmptyInlineSourceRef


class TestDuckDBEmpty(BaseTestEmpty):
    pass


class TestDuckDBEmptyInlineSourceRef(BaseTestEmptyInlineSourceRef):
    pass



from duckdb import DuckDBPyConnection

from dbt.adapters.duckdb.plugins import BasePlugin
from dbt.adapters.duckdb.utils import TargetConfig


def foo() -> int:
    return 1729


class Plugin(BasePlugin):
    def configure_connection(self, conn: DuckDBPyConnection):
        conn.create_function("foo", foo)

    def store(self, target_config: TargetConfig):
        assert target_config.config.get("key") == "value"



import pytest

from dbt.tests.adapter.unit_testing.test_types import BaseUnitTestingTypes
from dbt.tests.adapter.unit_testing.test_case_insensitivity import BaseUnitTestCaseInsensivity
from dbt.tests.adapter.unit_testing.test_invalid_input import BaseUnitTestInvalidInput


@pytest.mark.skip_profile("buenavista")
class TestUnitTestingTypesDuckDB(BaseUnitTestingTypes):
    @pytest.fixture
    def data_types(self):
        # sql_value, yaml_value
        return [
            ["1", "1"],
            ["2.0", "2.0"],
            ["'12345'", "12345"],
            ["'string'", "string"],
            ["true", "true"],
            ["DATE '2020-01-02'", "2020-01-02"],
            ["TIMESTAMP '2013-11-03 00:00:00-0'", "2013-11-03 00:00:00-0"],
            ["'2013-11-03 00:00:00-0'::TIMESTAMPTZ", "2013-11-03 00:00:00-0"],
            [
                "{'Alberta':'Edmonton','Manitoba':'Winnipeg'}",
                "{'Alberta':'Edmonton','Manitoba':'Winnipeg'}",
            ],
            ["ARRAY['a','b','c']", "['a','b','c']"],
            ["ARRAY[1,2,3]", "[1, 2, 3]"],
        ]


class TestUnitTestCaseInsensitivityDuckDB(BaseUnitTestCaseInsensivity):
    pass


class TestUnitTestInvalidInputDuckDB(BaseUnitTestInvalidInput):
    pass



"""
Test that the generate database name macro is case insensitive

See DuckDB docs: https://duckdb.org/docs/sql/dialect/keywords_and_identifiers.html

"Identifiers in DuckDB are always case-insensitive, similarly to PostgreSQL.
However, unlike PostgreSQL (and some other major SQL implementations), DuckDB also
treats quoted identifiers as case-insensitive."
"""
from urllib.parse import urlparse
import pytest

from dbt.tests.util import (
    run_dbt,
    check_result_nodes_by_name
)
from tests.functional.plugins.motherduck.fixtures import (
    models__gen_data_macro,
    macros__generate_database_name,
    macros__generate_schema_name,
    seeds__example_seed_csv,
)


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMacrosGenerateDatabaseName:
    @pytest.fixture(scope="class")
    def database_name(self, dbt_profile_target, request):
        return urlparse(dbt_profile_target["path"]).path + "_ducky_ducky"

    @pytest.fixture(autouse=True)
    def run_dbt_scope(self, project, database_name):
        project.run_sql(f"CREATE DATABASE IF NOT EXISTS {database_name}")
        yield
        project.run_sql(f"DROP DATABASE {database_name}")

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": seeds__example_seed_csv,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model.sql": models__gen_data_macro
        }

    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "db_name.sql": macros__generate_database_name,
            "schema_name.sql": macros__generate_schema_name
        }
    
    @staticmethod
    def gen_project_config_update(build_env, org_prefix):
        return {
            "config-version": 2,
            "vars": {
                "test": {
                    "build_env": build_env,
                    "org_prefix": org_prefix
                },
            },
            "macro-paths": ["macros"],
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return self.gen_project_config_update("ducky", "ducky")

    def test_dbname_macro(self, project):
        # seed command
        results = run_dbt(["seed"])
        assert len(results) == 1
        check_result_nodes_by_name(results, ["seed"])

        for _ in range(3):
            results = run_dbt(["run"])
            assert len(results) == 1
            check_result_nodes_by_name(results, ["model"])


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMacrosGenerateDatabaseNameUpperCase(TestMacrosGenerateDatabaseName):
    @pytest.fixture(scope="class")
    def database_name(self, dbt_profile_target, request):
        return urlparse(dbt_profile_target["path"]).path + "_ducky_ducky"

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return self.gen_project_config_update("DUCKY", "DUCKY")


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMacrosGenerateDatabaseNameLowerCase(TestMacrosGenerateDatabaseName):
    @pytest.fixture(scope="class")
    def database_name(self, dbt_profile_target, request):
        return urlparse(dbt_profile_target["path"]).path + "_DUCKY_DUCKY"

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return self.gen_project_config_update("ducky", "ducky")


@pytest.mark.skip_profile("buenavista", "file", "memory")
class TestMacrosGenerateDatabaseNameAllMixedCase(TestMacrosGenerateDatabaseName):
    @pytest.fixture(scope="class")
    def database_name(self, dbt_profile_target, request):
        return urlparse(dbt_profile_target["path"]).path + "_dUcKy_DUckY"

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return self.gen_project_config_update("DuCkY", "dUcKy")



import pytest

from dbt.tests.util import (
    check_relations_equal,
    run_dbt,
)

sources_schema_yml = """
version: 2
sources:
  - name: iceberg_source
    schema: main
    config:
      plugin: iceberg
      iceberg_table: "examples.{identifier}"
    tables:
      - name: nyc_taxi_locations
"""

models_source_model1_sql = """
    select * from {{ source('iceberg_source', 'nyc_taxi_locations') }}
"""


# Skipping this b/c it requires using my (@jwills) personal creds
# when testing it locally and also b/c I think there is something
# wrong with profiles_config_update since it can't be used in multiple
# tests in the same pytest session
@pytest.mark.skip
class TestIcebergPlugin:
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        config = {"catalog": "default"}
        if "path" not in dbt_profile_target:
            return {}
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target["path"],
                        "plugins": [
                            {"module": "iceberg", "config": config}
                        ],
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": sources_schema_yml,
            "source_model1.sql": models_source_model1_sql,
        }

    def test_iceberg_plugin(self, project):
        results = run_dbt()
        assert len(results) == 1

        res = project.run_sql("SELECT COUNT(1) FROM nyc_taxi_locations", fetch="one")
        assert res[0] == 265

        check_relations_equal(
            project.adapter,
            [
                "nyc_taxi_locations",
                "source_model1",
            ],
        )


from dbt.tests.adapter.caching.test_caching import (
    BaseCachingLowercaseModel,
    BaseCachingSelectedSchemaOnly,
)


class TestCachingLowerCaseModelDuckDB(BaseCachingLowercaseModel):
    pass


class TestCachingSelectedSchemaOnlyDuckDB(BaseCachingSelectedSchemaOnly):
    pass



import pytest
from unittest.mock import MagicMock
from unittest.mock import patch

import duckdb

from dbt.adapters.duckdb.credentials import Retries
from dbt.adapters.duckdb.environments import RetryableCursor

class TestRetryableCursor:

    @pytest.fixture
    def mock_cursor(self):
        return MagicMock()

    @pytest.fixture
    def mock_retries(self):
        return Retries(query_attempts=3)

    @pytest.fixture
    def retry_cursor(self, mock_cursor, mock_retries):
        return RetryableCursor(
            mock_cursor,
            mock_retries.query_attempts,
            mock_retries.retryable_exceptions)

    def test_successful_execute(self, mock_cursor, retry_cursor):
        """ Test that execute successfully runs the SQL query. """
        sql_query = "SELECT * FROM table"
        retry_cursor.execute(sql_query)
        mock_cursor.execute.assert_called_once_with(sql_query)

    def test_retry_on_failure(self, mock_cursor, retry_cursor):
        """ Test that execute retries the SQL query on failure. """
        mock_cursor.execute.side_effect = [duckdb.duckdb.IOException, None]
        sql_query = "SELECT * FROM table"
        retry_cursor.execute(sql_query)
        assert mock_cursor.execute.call_count == 2

    def test_no_retry_on_non_retryable_exception(self, mock_cursor, retry_cursor):
        """ Test that a non-retryable exception is not retried. """
        mock_cursor.execute.side_effect = ValueError
        sql_query = "SELECT * FROM table"
        with pytest.raises(ValueError):
            retry_cursor.execute(sql_query)
        mock_cursor.execute.assert_called_once_with(sql_query)

    def test_exponential_backoff(self, mock_cursor, retry_cursor):
        """ Test that exponential backoff is applied between retries. """
        mock_cursor.execute.side_effect = [duckdb.duckdb.IOException, duckdb.duckdb.IOException, None]
        sql_query = "SELECT * FROM table"

        with patch("time.sleep") as mock_sleep:
            retry_cursor.execute(sql_query)
            assert mock_sleep.call_count == 2
            mock_sleep.assert_any_call(1)
            mock_sleep.assert_any_call(2)



import os
import pytest
from dbt.tests.adapter.basic.files import (
    base_table_sql,
    model_base,
    schema_base_yml,
    seeds_base_csv,
)
from dbt.tests.util import (
    check_relation_types,
    check_relations_equal,
    check_result_nodes_by_name,
    relation_from_name,
    run_dbt,
)

config_write_csv_delim_options = """
  {{ config(materialized="external", format="csv", options={"delimiter": "|"}) }}
"""

config_write_codec_options = """
  {{ config(materialized="external", options={"codec": "zstd"}) }}
"""

config_write_partition_by_id = """
    {{ config(materialized="external", options={"partition_by": "id", "codec": "zstd"}) }}
"""

config_write_partition_by_id_name = """
    {{ config(materialized="external", options={"partition_by": "id, name"}) }}
"""

csv_delim_options_sql = config_write_csv_delim_options + model_base
write_codec_options = config_write_codec_options + model_base
config_write_partition_by_id_sql = config_write_partition_by_id + model_base
config_write_partition_by_id_name_sql = config_write_partition_by_id_name + model_base


class BaseExternalMaterializations:

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target, tmp_path_factory):
        extroot = str(tmp_path_factory.getbasetemp() / "write_options")
        os.mkdir(extroot)
        dbt_profile_target["external_root"] = extroot
        return dbt_profile_target

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model.sql": base_table_sql,
            "csv_delim_options.sql": csv_delim_options_sql,
            "write_codec_options.sql": write_codec_options,
            "config_write_partition_by_id.sql": config_write_partition_by_id_sql,
            "config_write_partition_by_id_name.sql": config_write_partition_by_id_name_sql,
            "schema.yml": schema_base_yml,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "base",
        }

    def test_base(self, project):

        # seed command
        results = run_dbt(["seed"])
        # seed result length
        assert len(results) == 1

        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 5

        # names exist in result nodes
        check_result_nodes_by_name(
            results,
            [
                "table_model",
                "csv_delim_options",
                "write_codec_options",
                "config_write_partition_by_id",
                "config_write_partition_by_id_name",
            ],
        )

        # check relation types
        expected = {
            "base": "table",
            "table_model": "table",
            "csv_delim_options": "view",
            "write_codec_options": "view",
            "config_write_partition_by_id": "view",
            "config_write_partition_by_id_name": "view",
        }
        check_relation_types(project.adapter, expected)

        # base table rowcount
        relation = relation_from_name(project.adapter, "base")
        result = project.run_sql(f"select count(*) as num_rows from {relation}", fetch="one")
        assert result[0] == 10

        # relations_equal
        check_relations_equal(
            project.adapter,
            [
                "base",
                "csv_delim_options",
                "write_codec_options",
                "config_write_partition_by_id",
                "config_write_partition_by_id_name",
            ],
        )

        # check relations in catalog
        catalog = run_dbt(["docs", "generate"])
        assert len(catalog.nodes) == 6
        assert len(catalog.sources) == 1


class TestExternalMaterializations(BaseExternalMaterializations):
    pass



from dbt.tests.adapter.aliases.test_aliases import BaseAliases, BaseAliasErrors, BaseSameAliasDifferentSchemas

class TestAliasesDuckDB(BaseAliases):
    pass

class TestAliasesErrorDuckDB(BaseAliasErrors):
    pass

class BaseSameALiasDifferentSchemasDuckDB(BaseSameAliasDifferentSchemas):
    pass



import pytest
from dbt.tests.adapter.python_model.test_python_model import (
    BasePythonIncrementalTests,
    BasePythonModelTests,
    basic_sql,
    m_1,
    schema_yml,
    second_sql,
)
from dbt.tests.util import run_dbt

basic_python_template = """
import pandas as pd

def model(dbt, _):
    dbt.config(
        materialized='table',
    )
    pdf = pd.DataFrame()
    df =  dbt.ref("my_sql_model")
    df2 = dbt.source('test_source', 'test_table')
    df = df.limit(2)
    return df{extension}
"""


class TestBasePythonModelDuckDBPyRelation(BasePythonModelTests):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": schema_yml,
            "my_sql_model.sql": basic_sql,
            "my_versioned_sql_model_v1.sql": basic_sql,
            "my_python_model.py": basic_python_template.format(extension=""),
            "second_sql_model.sql": second_sql,
        }


class TestBasePythonModelPandasDF(BasePythonModelTests):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": schema_yml,
            "my_sql_model.sql": basic_sql,
            "my_versioned_sql_model_v1.sql": basic_sql,
            "my_python_model.py": basic_python_template.format(extension=".df()"),
            "second_sql_model.sql": second_sql,
        }


incremental_python = """
def model(dbt, session):
    dbt.config(materialized="incremental", unique_key='id')
    df = dbt.ref("m_1")
    if dbt.is_incremental:
        # incremental runs should only apply to part of the data
        df = df.filter("id > 5")
    return df.df()
"""

# TODO(jwills): figure out why this one doesn't work; I think it's a test utils issue
@pytest.mark.skip_profile("buenavista")
class TestBasePythonIncremental(BasePythonIncrementalTests):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"+incremental_strategy": "delete+insert"}}

    @pytest.fixture(scope="class")
    def models(self):
        return {"m_1.sql": m_1, "incremental.py": incremental_python}


empty_upstream_model_python = """
def model(dbt, con):
    dbt.config(
        materialized='table',
    )
    return con.query("select 'a'::varchar as a, 0::boolean as b limit 0")
"""


class TestEmptyPythonModel:
    """
    This test ensures that Python models returning a DuckDBPyRelation are materialized
    with the correct schema, even when empty. I.e. ensure pyarrow is being used instead
    of pandas.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_model.py": empty_upstream_model_python,
        }

    def test_run(self, project):
        run_dbt(["run"])
        result = project.run_sql(
            """
            select column_name, data_type from system.information_schema.columns
            where table_name='upstream_model' order by column_name
            """,
            fetch="all",
        )
        assert result == [("a", "VARCHAR"), ("b", "BOOLEAN")]

temp_upstream_model_python = """
def model(dbt, con):
    dbt.config(
        materialized='table',
    )
    con.execute("create temp table t(a int)")
    return con.table("t")
"""


class TestTempTablePythonModel:
    """
    This test ensures that Python models returning a DuckDBPyRelation based
    on a temporary duckdb table can still be materialized
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "upstream_model.py": temp_upstream_model_python,
        }
    
    def test_run(self, project):
        run_dbt(["run"])

python_pyarrow_table_model = """
import pyarrow as pa

def model(dbt, con):
    return pa.Table.from_pydict({"a": [1,2,3]})
"""

python_pyarrow_dataset_model = """
import pyarrow as pa
import pyarrow.dataset as ds

def model(dbt, con):
    return ds.dataset(pa.Table.from_pydict({"b": [4, 5, 6]}))
"""


class TestMultiThreadedImports:
    """
    This test ensures that multiple pyarrow models can run concurrently with threads > 1
    and not suffer import issues.
    """

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_table1.py": python_pyarrow_table_model,
            "model_table2.py": python_pyarrow_table_model,
            "model_table3.py": python_pyarrow_table_model,
            "model_dataset1.py": python_pyarrow_dataset_model,
            "model_dataset2.py": python_pyarrow_dataset_model,
            "model_dataset3.py": python_pyarrow_dataset_model,
        }

    def test_run(self, project):
        run_dbt(["run"])



import pytest

from dbt.adapters.duckdb.column import DuckDBColumn

# Test cases for is_float method
@pytest.mark.parametrize("dtype, expected", [
    ("real", True),
    ("float", True),
    ("float4", True),
    ("float8", True),
    ("double", True),
    ("integer", False),
    ("string", False),
    ("bigint", False)
])
def test_is_float(dtype, expected):
    column = DuckDBColumn(column="float_test", dtype=dtype)
    assert column.is_float() == expected

# Test cases for is_integer method
@pytest.mark.parametrize("dtype, expected", [
    ("tinyint", True),
    ("smallint", True),
    ("integer", True),
    ("bigint", True),
    ("hugeint", True),
    ("utinyint", True),
    ("usmallint", True),
    ("uinteger", True),
    ("ubigint", True),
    ("int1", True),
    ("int2", True),
    ("int4", True),
    ("int8", True),
    ("short", True),
    ("int", True),
    ("signed", True),
    ("long", True),
    ("float", False),
    ("string", False),
    ("double", False)
])
def test_is_integer(dtype, expected):
    column = DuckDBColumn(column="integer_test", dtype=dtype)
    assert column.is_integer() == expected



import os
import pytest
import sqlite3

from dbt.tests.util import (
    check_relations_equal,
    run_dbt,
)

sqlalchemy_schema_yml = """
version: 2
sources:
  - name: sql_source
    schema: main
    config:
      plugin: sql
      save_mode: ignore
    tables:
      - name: tt1
        description: "My first SQLAlchemy table"
        config:
          query: "SELECT * FROM {identifier} WHERE id=:id"
          params:
            id: 1
      - name: tt2
        config:
          table: "test_table2"
"""


sqlalchemy1_sql = """
    select * from {{ source('sql_source', 'tt1') }}
"""
sqlalchemy2_sql = """
   {{ config(materialized='external', plugin='sql') }}
    select * from {{ source('sql_source', 'tt2') }}
"""
plugin_sql = """
    {{ config(materialized='external', plugin='cfp', key='value') }}
    select foo() as foo
"""


@pytest.mark.skip_profile("buenavista", "md")
class TestPlugins:
    @pytest.fixture(scope="class")
    def sqlite_test_db(self):
        path = "/tmp/satest.db"
        db = sqlite3.connect(path)
        cursor = db.cursor()
        cursor.execute("CREATE TABLE tt1 (id int, name text)")
        cursor.execute("INSERT INTO tt1 VALUES (1, 'John Doe')")
        cursor.execute("INSERT INTO tt1 VALUES (2, 'Jane Smith')")
        cursor.execute("CREATE TABLE test_table2 (a int, b int, c int)")
        cursor.execute("INSERT INTO test_table2 VALUES (1, 2, 3), (4, 5, 6)")
        cursor.close()
        db.commit()
        db.close()

        yield path

        # verify that the external plugin operation works to write to the db
        db = sqlite3.connect(path)
        cursor = db.cursor()
        res = cursor.execute("SELECT * FROM sqlalchemy2").fetchall()
        assert len(res) == 2
        assert res[0] == (1, 2, 3)
        assert res[1] == (4, 5, 6)
        cursor.close()
        db.close()

        os.unlink(path)

    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target, sqlite_test_db):
        sa_config = {"connection_url": f"sqlite:///{sqlite_test_db}"}
        plugins = [
            {"module": "sqlalchemy", "alias": "sql", "config": sa_config},
            {"module": "tests.create_function_plugin", "alias": "cfp"},
        ]

        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target.get("path", ":memory:"),
                        "plugins": plugins,
                        "retries": {"query_attempts": 2},
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self, test_data_path):
        return {
            "schema_sqlalchemy.yml": sqlalchemy_schema_yml,
            "sqlalchemy1.sql": sqlalchemy1_sql,
            "sqlalchemy2.sql": sqlalchemy2_sql,
            "foo.sql": plugin_sql,
        }

    def test_plugins(self, project):
        results = run_dbt()
        assert len(results) == 3

        res = project.run_sql("SELECT COUNT(1) FROM tt1", fetch="one")
        assert res[0] == 1
        check_relations_equal(
            project.adapter,
            [
                "tt1",
                "sqlalchemy1",
            ],
        )

        res = project.run_sql("SELECT COUNT(1) FROM tt2", fetch="one")
        assert res[0] == 2
        check_relations_equal(
            project.adapter,
            [
                "tt2",
                "sqlalchemy2",
            ],
        )

        res = project.run_sql("SELECT foo FROM foo", fetch="one")
        assert res[0] == 1729



from typing import Any
from typing import Dict
from urllib.parse import parse_qs
from urllib.parse import urlparse

from duckdb import DuckDBPyConnection

from . import BasePlugin
from dbt.adapters.duckdb.__version__ import version as __plugin_version__
from dbt.adapters.duckdb.credentials import DuckDBCredentials
from dbt.version import __version__

CUSTOM_USER_AGENT = "custom_user_agent"
MOTHERDUCK_EXT = "motherduck"
# MotherDuck config options, in order in which they need to be set
# (SaaS mode is last because it locks other config options)
MOTHERDUCK_CONFIG_OPTIONS = [
    "motherduck_token",
    "motherduck_attach_mode",
    "motherduck_saas_mode",
]


class Plugin(BasePlugin):
    def initialize(self, plugin_config: Dict[str, Any]):
        self._config = plugin_config

    @staticmethod
    def get_config_from_path(path):
        return {key: value[0] for key, value in parse_qs(urlparse(path).query).items()}

    @staticmethod
    def get_md_config_settings(config):
        # Get MotherDuck config settings
        md_config = {}
        for name in MOTHERDUCK_CONFIG_OPTIONS:
            for key in [
                name,
                name.replace("motherduck_", ""),
                name.upper(),
                name.replace("motherduck_", "").upper(),
            ]:
                if key in config:
                    md_config[name] = config[key]

        # Sort values (SaaS mode should be set last)
        return dict(
            sorted(
                md_config.items(),
                key=lambda x: MOTHERDUCK_CONFIG_OPTIONS.index(x[0]),
            )
        )

    def configure_connection(self, conn: DuckDBPyConnection):
        conn.load_extension(MOTHERDUCK_EXT)
        # If a MotherDuck database is in attachments,
        # set config options *before* attaching
        if self.creds is not None and self.creds.is_motherduck_attach:
            config = {}

            # add config options specified in the path
            for attachment in self.creds.motherduck_attach:
                config.update(self.get_config_from_path(attachment.path))

            # add config options specified via plugin config
            config.update(self._config)

            # add config options specified via settings
            if self.creds.settings is not None:
                config.update(self.creds.settings)

            # set MD config options and remove from settings
            for key, value in self.get_md_config_settings(config).items():
                conn.execute(f"SET {key} = '{value}'")
                if self.creds.settings is not None and key in self.creds.settings:
                    self.creds.settings.pop(key)

    def update_connection_config(self, creds: DuckDBCredentials, config: Dict[str, Any]):
        user_agent = f"dbt/{__version__} dbt-duckdb/{__plugin_version__}"
        settings: Dict[str, Any] = creds.settings or {}
        custom_user_agent = config.get(CUSTOM_USER_AGENT) or settings.pop(CUSTOM_USER_AGENT, None)
        if custom_user_agent:
            user_agent = f"{user_agent} {custom_user_agent}"
        config[CUSTOM_USER_AGENT] = user_agent

        # If a user specified MotherDuck config options via the plugin config,
        # pass it to the config kwarg in duckdb.connect.
        if not creds.is_motherduck_attach:
            config.update(self.get_md_config_settings(self._config))



from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence

from dbt.adapters.base.column import Column
from dbt.adapters.base.relation import BaseRelation
from dbt.adapters.contracts.relation import RelationConfig
# TODO
# from dbt.context.providers import RuntimeConfigObject


@dataclass
class SourceConfig:
    name: str
    identifier: str
    schema: str
    database: Optional[str]
    meta: Dict[str, Any]
    tags: List[str]

    def get(self, key, default=None):
        return self.meta.get(key, default)

    def __getitem__(self, key):
        return self.meta[key]

    def __contains__(self, key):
        return key in self.meta

    def table_name(self) -> str:
        if self.database:
            return ".".join([self.database, self.schema, self.identifier])
        else:
            return ".".join([self.schema, self.identifier])

    def as_dict(self) -> Dict[str, Any]:
        base = {
            "name": self.name,
            "identifier": self.identifier,
            "schema": self.schema,
            "database": self.database,
            "tags": self.tags,
        }
        base.update(self.meta)
        return base

    @classmethod
    def create_from_source(cls, source: RelationConfig) -> "SourceConfig":
        meta = source.meta.copy()
        # Use the config properties as well if they are present
        config_properties = source.config.extra if source.config else {}
        meta.update(config_properties)
        return SourceConfig(
            name=source.name,
            identifier=source.identifier,
            schema=source.schema,
            database=source.database,
            meta=meta,
            tags=source.tags or [],
        )


@dataclass
class TargetLocation:
    path: str
    format: str

    def as_dict(self) -> Dict[str, Any]:
        return {"path": self.path, "format": self.format}


@dataclass
class TargetConfig:
    relation: BaseRelation
    column_list: Sequence[Column]
    config: Any  # TODO
    location: Optional[TargetLocation] = None

    def as_dict(self) -> Dict[str, Any]:
        base = {
            "relation": self.relation.to_dict(),
            "column_list": [{"column": c.column, "dtype": c.dtype} for c in self.column_list],
            "config": self.config,
        }
        if self.location:
            base["location"] = self.location.as_dict()
        return base



import pytest

from dbt.tests.util import (
    check_relations_equal,
    run_dbt,
)

sources_schema_yml = """
version: 2
sources:
  - name: gsheet_source
    schema: main
    meta:
      plugin: gsheet
      title: "Josh's Test Spreadsheet"
    tables:
      - name: gsheet1
        description: "My first sheet"
      - name: gsheet2
        description: "The second sheet in the doc"
        meta:
          worksheet: "TwoSheet"
"""

models_source_model1_sql = """
    select * from {{ source('gsheet_source', 'gsheet1') }}
"""
models_source_model2_sql = """
    select * from {{ source('gsheet_source', 'gsheet2') }}
"""


# Skipping this b/c it requires using my (@jwills) personal creds
# when testing it locally and also b/c I think there is something
# wrong with profiles_config_update since it can't be used in multiple
# tests in the same pytest session
@pytest.mark.skip
class TestGSheetPlugin:
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        config = {"method": "oauth"}
        if "path" not in dbt_profile_target:
            return {}
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target["path"],
                        "plugins": [
                            {"module": "gsheet", "config": config}
                        ],
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self, test_data_path):
        return {
            "schema.yml": sources_schema_yml.format(test_data_path=test_data_path),
            "source_model1.sql": models_source_model1_sql,
            "source_model2.sql": models_source_model2_sql,
        }

    def test_gshseet_plugin(self, project):
        results = run_dbt()
        assert len(results) == 2

        check_relations_equal(
            project.adapter,
            [
                "gsheet1",
                "source_model1",
            ],
        )

        check_relations_equal(
            project.adapter,
            [
                "gsheet2",
                "source_model2",
            ],
        )



import pytest

from dbt.tests.adapter.persist_docs.test_persist_docs import (
    BasePersistDocs,
    BasePersistDocsColumnMissing,
    BasePersistDocsCommentOnQuotedColumn,
)

@pytest.mark.skip_profile("md")
class TestPersistDocs(BasePersistDocs):
    pass


@pytest.mark.skip_profile("md")
class TestPersistDocsColumnMissing(BasePersistDocsColumnMissing):
    pass


@pytest.mark.skip_profile("md")
class TestPersistDocsCommentOnQuotedColumn(BasePersistDocsCommentOnQuotedColumn):
    pass



#
# Models
#

models__gen_data_macro = """
select * from {{ ref("seed") }}
"""

#
# Macros
#

macros__generate_database_name = """
{% macro generate_database_name(custom_database_name=none, node=none) -%}
    {{ target.database | trim }}_{{ var("build_env") | trim }}_{{ var("org_prefix") | trim }}
{%- endmacro %}
"""


macros__generate_schema_name = """
{% macro generate_schema_name(custom_schema_name=none, node=none) -%}
    {{ target.schema | trim }}_{{ var("build_env") | trim }}_{{ var("org_prefix") | trim }}
{%- endmacro %}
"""

#
# Seeds
#

seeds__example_seed_csv = """a,b,c
1,2,3
4,5,6
7,8,9
"""



"""Unit test utility functions.

Note that all imports should be inside the functions to avoid import/mocking
issues.
"""
import os
from unittest import mock

from dbt.config.project import PartialProject


def normalize(path):
    """On windows, neither is enough on its own:

    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class Obj:
    which = "blah"
    single_threaded = False


def mock_connection(name):
    conn = mock.MagicMock()
    conn.name = name
    return conn


def profile_from_dict(profile, profile_name, cli_vars="{}"):
    from dbt.config import Profile
    from dbt.config.renderer import ProfileRenderer
    from dbt.config.utils import parse_cli_vars

    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = ProfileRenderer(cli_vars)
    return Profile.from_raw_profile_info(
        profile,
        profile_name,
        renderer,
    )


def project_from_dict(project, profile, packages=None, selectors=None, cli_vars="{}"):
    from dbt.config.renderer import DbtProjectYamlRenderer
    from dbt.config.utils import parse_cli_vars

    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = DbtProjectYamlRenderer(profile, cli_vars)

    project_root = project.pop("project-root", os.getcwd())

    partial = PartialProject.from_dicts(
        project_root=project_root,
        project_dict=project,
        packages_dict=packages,
        selectors_dict=selectors,
    )
    return partial.render(renderer)


def config_from_parts_or_dicts(project, profile, packages=None, selectors=None, cli_vars="{}"):
    from copy import deepcopy

    from dbt.config import Profile, Project, RuntimeConfig

    if isinstance(project, Project):
        profile_name = project.profile_name
    else:
        profile_name = project.get("profile")

    if not isinstance(profile, Profile):
        profile = profile_from_dict(
            deepcopy(profile),
            profile_name,
            cli_vars,
        )

    if not isinstance(project, Project):
        project = project_from_dict(
            deepcopy(project),
            profile,
            packages,
            selectors,
            cli_vars,
        )

    args = Obj()
    args.vars = cli_vars
    args.profile_dir = "/dev/null"
    return RuntimeConfig.from_parts(project=project, profile=profile, args=args)


def inject_plugin(plugin):
    from dbt.adapters.factory import FACTORY

    key = plugin.adapter.type()
    FACTORY.plugins[key] = plugin


def inject_adapter(value, plugin):
    """Inject the given adapter into the adapter factory, so your hand-crafted
    artisanal adapter will be available from get_adapter() as if dbt loaded it.
    """
    inject_plugin(plugin)
    from dbt.adapters.factory import FACTORY

    key = value.type()
    FACTORY.adapters[key] = value


def generate_name_macros(package):
    from dbt.contracts.graph.parsed import ParsedMacro
    from dbt.node_types import NodeType

    name_sql = {}
    for component in ("database", "schema", "alias"):
        if component == "alias":
            source = "node.name"
        else:
            source = f"target.{component}"
        name = f"generate_{component}_name"
        sql = f"{{% macro {name}(value, node) %}} {{% if value %}} {{{{ value }}}} {{% else %}} {{{{ {source} }}}} {{% endif %}} {{% endmacro %}}"
        name_sql[name] = sql

    all_sql = "\n".join(name_sql.values())
    for name, sql in name_sql.items():
        pm = ParsedMacro(
            name=name,
            resource_type=NodeType.Macro,
            unique_id=f"macro.{package}.{name}",
            package_name=package,
            original_file_path=normalize("macros/macro.sql"),
            root_path="./dbt_modules/root",
            path=normalize("macros/macro.sql"),
            raw_sql=all_sql,
            macro_sql=sql,
        )
        yield pm



import pytest

from dbt.tests.adapter.utils.test_any_value import BaseAnyValue
from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct
from dbt.tests.adapter.utils.test_bool_or import BaseBoolOr
from dbt.tests.adapter.utils.test_cast_bool_to_text import BaseCastBoolToText
from dbt.tests.adapter.utils.test_concat import BaseConcat
from dbt.tests.adapter.utils.test_current_timestamp import BaseCurrentTimestampNaive
from dbt.tests.adapter.utils.test_date_trunc import BaseDateTrunc
from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd
from dbt.tests.adapter.utils.test_datediff import BaseDateDiff
from dbt.tests.adapter.utils.test_date_spine import BaseDateSpine
from dbt.tests.adapter.utils.test_escape_single_quotes import (
    BaseEscapeSingleQuotesQuote,
)
from dbt.tests.adapter.utils.test_except import BaseExcept
from dbt.tests.adapter.utils.test_generate_series import BaseGenerateSeries
from dbt.tests.adapter.utils.test_get_intervals_between import BaseGetIntervalsBetween
from dbt.tests.adapter.utils.test_get_powers_of_two import BaseGetPowersOfTwo
from dbt.tests.adapter.utils.test_hash import BaseHash
from dbt.tests.adapter.utils.test_intersect import BaseIntersect
from dbt.tests.adapter.utils.test_last_day import BaseLastDay
from dbt.tests.adapter.utils.test_length import BaseLength
from dbt.tests.adapter.utils.test_listagg import BaseListagg
from dbt.tests.adapter.utils.test_position import BasePosition
from dbt.tests.adapter.utils.test_replace import BaseReplace
from dbt.tests.adapter.utils.test_right import BaseRight
from dbt.tests.adapter.utils.test_safe_cast import BaseSafeCast
from dbt.tests.adapter.utils.test_split_part import BaseSplitPart
from dbt.tests.adapter.utils.test_string_literal import BaseStringLiteral


class TestAnyValue(BaseAnyValue):
    pass


class TestBoolOr(BaseBoolOr):
    pass


class TestCastBoolToText(BaseCastBoolToText):
    pass


class TestConcat(BaseConcat):
    pass


class TestDateAdd(BaseDateAdd):
    pass


class TestDateDiff(BaseDateDiff):
    pass


# Skipping this b/c the upstream utils test
# is irritatingly adapter-specific at the moment
@pytest.mark.skip
class TestDateSpine(BaseDateSpine):
    pass

class TestDateTrunc(BaseDateTrunc):
    pass


class TestEscapeSingleQuotes(BaseEscapeSingleQuotesQuote):
    pass


class TestGenerateSeries(BaseGenerateSeries):
    pass


# Skipping this b/c the upstream utils test
# is irritatingly adapter-specific at the moment
@pytest.mark.skip
class TestGetIntervalsBetween(BaseGetIntervalsBetween):
    pass


class TestGetPowersOfTwo(BaseGetPowersOfTwo):
    pass


class TestExcept(BaseExcept):
    pass


class TestHash(BaseHash):
    pass


class TestIntersect(BaseIntersect):
    pass


class TestLastDay(BaseLastDay):
    pass


class TestLength(BaseLength):
    pass


# NOTE: list_agg relies on an ORDER BY construct MD does not yet support
@pytest.mark.skip_profile("md")
class TestListagg(BaseListagg):
    pass


class TestPosition(BasePosition):
    pass


class TestReplace(BaseReplace):
    pass


class TestRight(BaseRight):
    pass


class TestSafeCast(BaseSafeCast):
    pass


class TestSplitPart(BaseSplitPart):
    pass


class TestStringLiteral(BaseStringLiteral):
    pass


class TestArrayAppend(BaseArrayAppend):
    pass


class TestArrayConcat(BaseArrayConcat):
    pass


class TestArrayConstruct(BaseArrayConstruct):
    pass


# Skipping this while we sort out what the right default is
# here: https://github.com/duckdb/duckdb/issues/7934
@pytest.mark.skip
class TestCurrentTimestamp(BaseCurrentTimestampNaive):
    pass



import pytest

from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import (
    BaseSingularTestsEphemeral,
)
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_incremental import BaseIncremental
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp
from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod
from dbt.tests.adapter.basic.test_validate_connection import BaseValidateConnection
from dbt.tests.adapter.basic.test_docs_generate import (
    BaseDocsGenerate,
    BaseDocsGenReferences,
)
from dbt.tests.adapter.basic.expected_catalog import (
    base_expected_catalog,
    no_stats,
    expected_references_catalog,
)


class TestSimpleMaterializationsDuckDB(BaseSimpleMaterializations):
    pass


class TestSingularTestsDuckDB(BaseSingularTests):
    pass


class TestSingularTestsEphemeralDuckDB(BaseSingularTestsEphemeral):
    pass


class TestEmptyDuckDB(BaseEmpty):
    pass


class TestEphemeralDuckDB(BaseEphemeral):
    pass


class TestIncrementalDuckDB(BaseIncremental):
    pass


class TestGenericTestsDuckDB(BaseGenericTests):
    pass


class TestSnapshotCheckColsDuckDB(BaseSnapshotCheckCols):
    pass


class TestSnapshotTimestampDuckDB(BaseSnapshotTimestamp):
    pass


class TestBaseAdapterMethodDuckDB(BaseAdapterMethod):
    pass


class TestValidateConnectionDuckDB(BaseValidateConnection):
    pass


class TestDocsGenerateDuckDB(BaseDocsGenerate):
    @pytest.fixture(scope="class")
    def expected_catalog(self, project):
        return base_expected_catalog(
            project,
            role=None,
            id_type="INTEGER",
            text_type="VARCHAR",
            time_type="TIMESTAMP",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=no_stats(),
        )


class TestDocsGenReferencesDuckDB(BaseDocsGenReferences):
    @pytest.fixture(scope="class")
    def expected_catalog(self, project):
        return expected_references_catalog(
            project,
            role=None,
            id_type="INTEGER",
            text_type="VARCHAR",
            time_type="TIMESTAMP",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=no_stats(),
            bigint_type="BIGINT",
        )



import pandas
import pytest

from dbt.tests.util import (
    check_relations_equal,
    run_dbt,
)

schema_yml = """
version: 2
sources:
  - name: excel_source
    schema: main
    meta:
      plugin: excel
    tables:
      - name: excel_file
        description: "An excel file"
        meta:
          external_location: "{test_data_path}/excel_file.xlsx"
"""

plugins = [
    {
        "module": "excel",
        "config": {
            "output": {
                "engine": "openpyxl",
                "file": "/tmp/excel_file_out.xlsx",
                "lazy_close": False
            }
        }
    },
]

model_sql = """
    {{ config(materialized='external', plugin='excel') }}
    select * from {{ source('excel_source', 'excel_file') }}
"""


class TestExcelPlugin:
    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target.get("path", ":memory:"),
                        "plugins": plugins,
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self, test_data_path):
        return {
            "schema_excel.yml": schema_yml.format(test_data_path=test_data_path),
            "excel_read_write.sql": model_sql,
        }

    def test_excel_plugin(self, project):
        results = run_dbt()
        assert len(results) == 1

        res = project.run_sql("SELECT COUNT(1) FROM excel_file", fetch="one")
        assert res[0] == 9

        df = pandas.read_excel('/tmp/excel_file_out.xlsx')
        assert df.shape[0] == 9
        assert df['First Name'].iloc[0] == 'Dulce'

        check_relations_equal(
            project.adapter,
            [
                "excel_file",
                "excel_read_write",
            ],
        )





from typing import Any
from typing import Dict

from deltalake import DeltaTable

from . import BasePlugin
from ..utils import SourceConfig


class Plugin(BasePlugin):
    def initialize(self, config: Dict[str, Any]):
        pass

    def configure_cursor(self, cursor):
        pass

    def load(self, source_config: SourceConfig):
        if "delta_table_path" not in source_config:
            raise Exception("'delta_table_path' is a required argument for the delta table!")

        table_path = source_config["delta_table_path"]
        storage_options = source_config.get("storage_options", None)

        if storage_options:
            dt = DeltaTable(table_path, storage_options=storage_options)
        else:
            dt = DeltaTable(table_path)

        # delta attributes
        as_of_version = source_config.get("as_of_version", None)
        as_of_datetime = source_config.get("as_of_datetime", None)

        if as_of_version:
            dt.load_as_version(as_of_version)

        if as_of_datetime:
            dt.load_as_version(as_of_datetime)

        df = dt.to_pyarrow_dataset()

        return df

    def default_materialization(self):
        return "view"


# Future
# TODO add databricks catalog



from unittest.mock import call

import pytest

from dbt.adapters.base.column import Column
from dbt.adapters.duckdb.plugins.glue import create_or_update_table


class TestGlue:
    @pytest.fixture
    def columns(self):
        return [
            Column(column="bigint", dtype="BIGINT"),
            Column(column="int8", dtype="INT8"),
            Column(column="long", dtype="LONG"),
            Column(column="boolean", dtype="BOOLEAN"),
            Column(column="logical", dtype="LOGICAL"),
            Column(column="blob", dtype="BLOB"),
            Column(column="bytea", dtype="BYTEA"),
            Column(column="varbinary", dtype="VARBINARY"),
            Column(column="date", dtype="DATE"),
            Column(column="double", dtype="DOUBLE"),
            Column(column="float8", dtype="FLOAT8"),
            Column(column="numeric", dtype="NUMERIC"),
            Column(column="decimal", dtype="DECIMAL"),
            Column(column="decimal_3_2", dtype="DECIMAL"),
            Column(column="integer", dtype="INTEGER"),
            Column(column="int4", dtype="INT4"),
            Column(column="int", dtype="INT"),
            Column(column="signed", dtype="SIGNED"),
            Column(column="real", dtype="REAL"),
            Column(column="float4", dtype="FLOAT4"),
            Column(column="float", dtype="FLOAT"),
            Column(column="smallint", dtype="SMALLINT"),
            Column(column="int2", dtype="INT2"),
            Column(column="short", dtype="SHORT"),
            Column(column="timestamp", dtype="TIMESTAMP"),
            Column(column="datetime", dtype="DATETIME"),
            Column(column="timestamptz", dtype="TIMESTAMPTZ"),
            Column(column="tinyint", dtype="TINYINT"),
            Column(column="int1", dtype="INT1"),
            Column(column="uinteger", dtype="UINTEGER"),
            Column(column="usmallint", dtype="USMALLINT"),
            Column(column="utinyint", dtype="UTINYINT"),
            Column(column="uuid", dtype="UUID"),
            Column(column="varchar", dtype="VARCHAR"),
            Column(column="char", dtype="CHAR"),
            Column(column="bpchar", dtype="BPCHAR"),
            Column(column="text", dtype="TEXT"),
            Column(column="string", dtype="STRING"),
        ]

    def test_create_glue_table(self, mocker, columns):
        client = mocker.Mock()
        client.get_table.return_value = None
        create_or_update_table(
            client,
            database="test",
            table="test",
            column_list=columns,
            s3_path="s3://test/aaa.parquet",
            file_format="parquet",
            delimiter=",",
        )

        client.assert_has_calls(
            [
                call.get_table(DatabaseName="test", Name="test"),
                call.create_table(
                    DatabaseName="test",
                    TableInput={
                        "Name": "test",
                        "TableType": "EXTERNAL_TABLE",
                        "Parameters": {"classification": "parquet"},
                        "StorageDescriptor": {
                            "Columns": [
                                {"Name": "bigint", "Type": "bigint"},
                                {"Name": "int8", "Type": "bigint"},
                                {"Name": "long", "Type": "bigint"},
                                {"Name": "boolean", "Type": "boolean"},
                                {"Name": "logical", "Type": "boolean"},
                                {"Name": "blob", "Type": "binary"},
                                {"Name": "bytea", "Type": "binary"},
                                {"Name": "varbinary", "Type": "binary"},
                                {"Name": "date", "Type": "date"},
                                {"Name": "double", "Type": "double"},
                                {"Name": "float8", "Type": "double"},
                                {"Name": "numeric", "Type": "double"},
                                {"Name": "decimal", "Type": "double"},
                                {"Name": "decimal_3_2", "Type": "double"},
                                {"Name": "integer", "Type": "int"},
                                {"Name": "int4", "Type": "int"},
                                {"Name": "int", "Type": "int"},
                                {"Name": "signed", "Type": "int"},
                                {"Name": "real", "Type": "float"},
                                {"Name": "float4", "Type": "float"},
                                {"Name": "float", "Type": "float"},
                                {"Name": "smallint", "Type": "smallint"},
                                {"Name": "int2", "Type": "smallint"},
                                {"Name": "short", "Type": "smallint"},
                                {"Name": "timestamp", "Type": "timestamp"},
                                {"Name": "datetime", "Type": "timestamp"},
                                {"Name": "timestamptz", "Type": "timestamp"},
                                {"Name": "tinyint", "Type": "tinyint"},
                                {"Name": "int1", "Type": "tinyint"},
                                {"Name": "uinteger", "Type": "bigint"},
                                {"Name": "usmallint", "Type": "int"},
                                {"Name": "utinyint", "Type": "smallint"},
                                {"Name": "uuid", "Type": "string"},
                                {"Name": "varchar", "Type": "string"},
                                {"Name": "char", "Type": "string"},
                                {"Name": "bpchar", "Type": "string"},
                                {"Name": "text", "Type": "string"},
                                {"Name": "string", "Type": "string"},
                            ],
                            "Location": "s3://test",
                            "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
                            "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                            "Compressed": False,
                            "NumberOfBuckets": -1,
                            "SerdeInfo": {
                                "SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
                                "Parameters": {"serialization.format": "1"},
                            },
                            "BucketColumns": [],
                            "StoredAsSubDirectories": False,
                            "SortColumns": [],
                        },
                    },
                ),
            ]
        )

    def test_update_glue_table(self, mocker, columns):
        client = mocker.Mock()
        client.get_table.return_value = {
            "Table": {
                "Name": "test",
                "TableType": "EXTERNAL_TABLE",
                "Parameters": {"classification": "parquet"},
                "StorageDescriptor": {
                    "Columns": [
                        {"Name": "n_legs", "Type": "bigint"},
                    ],
                    "Location": "test",
                    "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
                    "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                    "Compressed": False,
                    "NumberOfBuckets": -1,
                    "SerdeInfo": {
                        "SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
                        "Parameters": {"serialization.format": "1"},
                    },
                    "BucketColumns": [],
                    "StoredAsSubDirectories": False,
                    "SortColumns": [],
                },
            },
        }
        create_or_update_table(
            client,
            database="test",
            table="test",
            column_list=columns,
            s3_path="s3://test/aaa.parquet",
            file_format="parquet",
            delimiter=",",
        )
        client.assert_has_calls(
            [
                call.get_table(DatabaseName="test", Name="test"),
                call.update_table(
                    DatabaseName="test",
                    TableInput={
                        "Name": "test",
                        "TableType": "EXTERNAL_TABLE",
                        "Parameters": {"classification": "parquet"},
                        "StorageDescriptor": {
                            "Columns": [
                                {"Name": "bigint", "Type": "bigint"},
                                {"Name": "int8", "Type": "bigint"},
                                {"Name": "long", "Type": "bigint"},
                                {"Name": "boolean", "Type": "boolean"},
                                {"Name": "logical", "Type": "boolean"},
                                {"Name": "blob", "Type": "binary"},
                                {"Name": "bytea", "Type": "binary"},
                                {"Name": "varbinary", "Type": "binary"},
                                {"Name": "date", "Type": "date"},
                                {"Name": "double", "Type": "double"},
                                {"Name": "float8", "Type": "double"},
                                {"Name": "numeric", "Type": "double"},
                                {"Name": "decimal", "Type": "double"},
                                {"Name": "decimal_3_2", "Type": "double"},
                                {"Name": "integer", "Type": "int"},
                                {"Name": "int4", "Type": "int"},
                                {"Name": "int", "Type": "int"},
                                {"Name": "signed", "Type": "int"},
                                {"Name": "real", "Type": "float"},
                                {"Name": "float4", "Type": "float"},
                                {"Name": "float", "Type": "float"},
                                {"Name": "smallint", "Type": "smallint"},
                                {"Name": "int2", "Type": "smallint"},
                                {"Name": "short", "Type": "smallint"},
                                {"Name": "timestamp", "Type": "timestamp"},
                                {"Name": "datetime", "Type": "timestamp"},
                                {"Name": "timestamptz", "Type": "timestamp"},
                                {"Name": "tinyint", "Type": "tinyint"},
                                {"Name": "int1", "Type": "tinyint"},
                                {"Name": "uinteger", "Type": "bigint"},
                                {"Name": "usmallint", "Type": "int"},
                                {"Name": "utinyint", "Type": "smallint"},
                                {"Name": "uuid", "Type": "string"},
                                {"Name": "varchar", "Type": "string"},
                                {"Name": "char", "Type": "string"},
                                {"Name": "bpchar", "Type": "string"},
                                {"Name": "text", "Type": "string"},
                                {"Name": "string", "Type": "string"},
                            ],
                            "Location": "s3://test",
                            "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
                            "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                            "Compressed": False,
                            "NumberOfBuckets": -1,
                            "SerdeInfo": {
                                "SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
                                "Parameters": {"serialization.format": "1"},
                            },
                            "BucketColumns": [],
                            "StoredAsSubDirectories": False,
                            "SortColumns": [],
                        },
                    },
                ),
            ]
        )

    def test_without_update_glue_table(self, mocker, columns):
        client = mocker.Mock()
        client.get_table.return_value = {
            "Table": {
                "Name": "test",
                "TableType": "EXTERNAL_TABLE",
                "Parameters": {
                    "classification": "parquet",
                },
                "StorageDescriptor": {
                    "Columns": [
                        {"Name": "bigint", "Type": "bigint"},
                        {"Name": "int8", "Type": "bigint"},
                        {"Name": "long", "Type": "bigint"},
                        {"Name": "boolean", "Type": "boolean"},
                        {"Name": "logical", "Type": "boolean"},
                        {"Name": "blob", "Type": "binary"},
                        {"Name": "bytea", "Type": "binary"},
                        {"Name": "varbinary", "Type": "binary"},
                        {"Name": "date", "Type": "date"},
                        {"Name": "double", "Type": "double"},
                        {"Name": "float8", "Type": "double"},
                        {"Name": "numeric", "Type": "double"},
                        {"Name": "decimal", "Type": "double"},
                        {"Name": "decimal_3_2", "Type": "double"},
                        {"Name": "integer", "Type": "int"},
                        {"Name": "int4", "Type": "int"},
                        {"Name": "int", "Type": "int"},
                        {"Name": "signed", "Type": "int"},
                        {"Name": "real", "Type": "float"},
                        {"Name": "float4", "Type": "float"},
                        {"Name": "float", "Type": "float"},
                        {"Name": "smallint", "Type": "smallint"},
                        {"Name": "int2", "Type": "smallint"},
                        {"Name": "short", "Type": "smallint"},
                        {"Name": "timestamp", "Type": "timestamp"},
                        {"Name": "datetime", "Type": "timestamp"},
                        {"Name": "timestamptz", "Type": "timestamp"},
                        {"Name": "tinyint", "Type": "tinyint"},
                        {"Name": "int1", "Type": "tinyint"},
                        {"Name": "uinteger", "Type": "bigint"},
                        {"Name": "usmallint", "Type": "int"},
                        {"Name": "utinyint", "Type": "smallint"},
                        {"Name": "uuid", "Type": "string"},
                        {"Name": "varchar", "Type": "string"},
                        {"Name": "char", "Type": "string"},
                        {"Name": "bpchar", "Type": "string"},
                        {"Name": "text", "Type": "string"},
                        {"Name": "string", "Type": "string"},
                    ],
                    "Location": "test",
                    "InputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat",
                    "OutputFormat": "org.apache.hadoop.hive.ql.io.parquet.MapredParquetOutputFormat",
                    "Compressed": True,
                    "NumberOfBuckets": -1,
                    "SerdeInfo": {
                        "SerializationLibrary": "org.apache.hadoop.hive.ql.io.parquet.serde.ParquetHiveSerDe",
                        "Parameters": {"serialization.format": "1"},
                    },
                    "BucketColumns": [],
                    "StoredAsSubDirectories": False,
                    "SortColumns": [],
                },
            },
        }
        create_or_update_table(
            client,
            database="test",
            table="test",
            column_list=columns,
            s3_path="test",
            file_format="parquet",
            delimiter=",",
        )
        assert len(client.mock_calls) == 1
        client.assert_has_calls([call.get_table(DatabaseName="test", Name="test")])



import os

import pytest
from dbt.tests.util import run_dbt

sources_schema_yml = """version: 2
sources:
  - name: external_source
    config:
      external_location: "/tmp/{name}_{extra}.csv"
    tables:
      - name: seeds_source
        description: "A source table"
        config:
          extra: 'something'
        columns:
          - name: id
            description: "An id"
            tests:
              - unique
              - not_null
      - name: seeds_ost
        identifier: "seeds_other_source_table"
        config:
          external_location: "read_csv_auto('/tmp/%(identifier)s.csv')"
          formatter: oldstyle
      - name: seeds_other_source_table
        config:
          external_location: "read_csv_auto('/tmp/${name}.csv')"
          formatter: template
"""

models_source_model_sql = """select * from {{ source('external_source', 'seeds_source') }}
"""

models_multi_source_model_sql = """select s.* from {{ source('external_source', 'seeds_source') }} s
  inner join {{ source('external_source', 'seeds_ost') }} oldstyle USING (id)
  inner join {{ source('external_source', 'seeds_other_source_table') }} tmpl USING (id)
"""


class TestExternalSources:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "schema.yml": sources_schema_yml,
            "source_model.sql": models_source_model_sql,
            "multi_source_model.sql": models_multi_source_model_sql,
        }

    @pytest.fixture(scope="class")
    def seeds_source_file(self):
        with open("/tmp/seeds_source_something.csv", "w") as f:
            f.write("id,a,b\n1,2,3\n4,5,6\n7,8,9")
        yield
        os.unlink("/tmp/seeds_source_something.csv")

    @pytest.fixture(scope="class")
    def ost_file(self):
        with open("/tmp/seeds_other_source_table.csv", "w") as f:
            f.write("id,c,d\n1,2,3\n4,5,6\n7,8,9")
        yield
        os.unlink("/tmp/seeds_other_source_table.csv")

    def test_external_sources(self, seeds_source_file, ost_file, project):
        results = run_dbt(["run"])
        assert len(results) == 2
        test_results = run_dbt(["test"])
        assert len(test_results) == 2



import duckdb
import pytest
from unittest import mock

from botocore.credentials import Credentials

from dbt.adapters.duckdb.credentials import Attachment, DuckDBCredentials


def test_load_basic_settings():
    settings = {
        "access_mode": "read_only",
        "log_query_path": "/path/to/log",
    }
    creds = DuckDBCredentials(settings=settings)
    assert creds.settings == settings


def test_add_secret_with_empty_name():
    creds = DuckDBCredentials(
        secrets=[
            dict(
                type="s3",
                name="",
                key_id="abc",
                secret="xyz",
                region="us-west-2"
            )
        ]
    )
    assert len(creds.secrets) == 1
    assert creds._secrets[0].type == "s3"
    assert creds._secrets[0].secret_kwargs.get("key_id") == "abc"
    assert creds._secrets[0].secret_kwargs.get("secret") == "xyz"
    assert creds._secrets[0].secret_kwargs.get("region") == "us-west-2"

    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE SECRET (
    type s3,
    key_id 'abc',
    secret 'xyz',
    region 'us-west-2'
)"""


def test_add_secret_with_name():
    creds = DuckDBCredentials.from_dict(
        dict(secrets=[
            dict(
                type="s3",
                name="my_secret",
                key_id="abc",
                secret="xyz",
                region="us-west-2",
                scope="s3://my-bucket"
            )
        ])
    )
    assert len(creds._secrets) == 1
    assert creds._secrets[0].type == "s3"
    assert creds._secrets[0].secret_kwargs.get("key_id") == "abc"
    assert creds._secrets[0].secret_kwargs.get("secret") == "xyz"
    assert creds._secrets[0].secret_kwargs.get("region") == "us-west-2"
    assert creds._secrets[0].scope == "s3://my-bucket"

    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE OR REPLACE SECRET my_secret (
    type s3,
    scope 's3://my-bucket',
    key_id 'abc',
    secret 'xyz',
    region 'us-west-2'
)"""


def test_add_unsupported_secret():
    creds = DuckDBCredentials(
        secrets=[
            dict(
                type="scrooge_mcduck",
                name="money"
            )
        ]
    )
    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE OR REPLACE SECRET money (
    type scrooge_mcduck
)"""
    with pytest.raises(duckdb.InvalidInputException) as e:
        duckdb.sql(sql)
        assert "Secret type 'scrooge_mcduck' not found" in str(e)


@pytest.mark.skip_profile("nightly")
def test_add_unsupported_secret_param():
    creds = DuckDBCredentials(
        secrets=[
            dict(
                type="s3",
                password="secret"
            )
        ]
    )
    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE OR REPLACE SECRET _dbt_secret_1 (
    type s3,
    password 'secret'
)"""
    with pytest.raises(duckdb.BinderException) as e:
        duckdb.sql(sql)
        msg = "Unknown parameter 'password' for secret type 's3' with default provider 'config'"
        assert msg in str(e)


def test_add_azure_secret():
    creds = DuckDBCredentials(
        secrets=[
            dict(
                type="azure",
                name="",
                provider="service_principal",
                tenant_id="abc",
                client_id="xyz",
                client_certificate_path="foo\\bar\\baz.pem",
                account_name="123"
            )
        ]
    )
    assert len(creds.secrets) == 1
    assert creds._secrets[0].type == "azure"
    assert creds._secrets[0].secret_kwargs.get("tenant_id") == "abc"
    assert creds._secrets[0].secret_kwargs.get("client_id") == "xyz"
    assert creds._secrets[0].secret_kwargs.get("client_certificate_path") == "foo\\bar\\baz.pem"
    assert creds._secrets[0].secret_kwargs.get("account_name") == "123"

    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE SECRET (
    type azure,
    provider service_principal,
    tenant_id 'abc',
    client_id 'xyz',
    client_certificate_path 'foo\\bar\\baz.pem',
    account_name '123'
)"""


def test_add_hf_secret():
    creds = DuckDBCredentials(
        secrets=[
            dict(
                type="huggingface",
                name="",
                token="abc"
            )
        ]
    )
    assert len(creds.secrets) == 1
    assert creds._secrets[0].type == "huggingface"
    assert creds._secrets[0].secret_kwargs.get("token") == "abc"

    sql = creds.secrets_sql()[0]
    assert sql == \
"""CREATE SECRET (
    type huggingface,
    token 'abc'
)"""


@mock.patch("boto3.session.Session")
def test_load_aws_creds(mock_session_class):
    mock_session_object = mock.Mock()
    mock_client = mock.Mock()

    mock_session_object.get_credentials.return_value = Credentials(
        "access_key", "secret_key", "token"
    )
    mock_session_object.client.return_value = mock_client
    mock_session_class.return_value = mock_session_object
    mock_client.get_caller_identity.return_value = {}

    creds = DuckDBCredentials(use_credential_provider="aws")
    assert len(creds.secrets) == 1
    assert creds._secrets[0].type == "s3"
    assert creds._secrets[0].provider == "credential_chain"


def test_attachments():
    creds = DuckDBCredentials()
    creds.attach = [
        {"path": "/tmp/f1234.db"},
        {"path": "/tmp/g1234.db", "alias": "g"},
        {"path": "/tmp/h5678.db", "read_only": 1},
        {"path": "/tmp/i9101.db", "type": "sqlite"},
        {"path": "/tmp/jklm.db", "alias": "jk", "read_only": 1, "type": "sqlite"},
    ]

    expected_sql = [
        "ATTACH '/tmp/f1234.db'",
        "ATTACH '/tmp/g1234.db' AS g",
        "ATTACH '/tmp/h5678.db' (READ_ONLY)",
        "ATTACH '/tmp/i9101.db' (TYPE sqlite)",
        "ATTACH '/tmp/jklm.db' AS jk (TYPE sqlite, READ_ONLY)",
    ]

    for i, a in enumerate(creds.attach):
        attachment = Attachment(**a)
        assert expected_sql[i] == attachment.to_sql()


def test_infer_database_name_from_path():
    payload = {}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "memory"

    payload = {"path": "local.duckdb"}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "local"

    payload = {"path": "/tmp/f1234.db"}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "f1234"

    payload = {"path": "md:?token=abc123"}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "my_db"

    payload = {"path": "md:jaffle_shop?token=abc123"}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "jaffle_shop"

    payload = {"database": "memory"}
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "memory"

    payload = {
        "database": "remote",
        "remote": {"host": "localhost", "port": 5433, "user": "test"},
    }
    creds = DuckDBCredentials.from_dict(payload)
    assert creds.database == "remote"



import abc
import importlib.util
import os
import sys
import tempfile
import time
from typing import Dict
from typing import List
from typing import Optional

import duckdb
from dbt_common.exceptions import DbtRuntimeError

from ..credentials import DuckDBCredentials
from ..credentials import Extension
from ..plugins import BasePlugin
from ..utils import SourceConfig
from ..utils import TargetConfig
from dbt.adapters.contracts.connection import AdapterResponse
from dbt.adapters.contracts.connection import Connection


def _ensure_event_loop():
    """
    Ensures the current thread has an event loop defined, and creates one if necessary.
    """
    import asyncio

    try:
        # Check that the current thread has an event loop attached
        loop = asyncio.get_event_loop()
    except RuntimeError:
        # If the current thread doesn't have an event loop, create one
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)


class RetryableCursor:
    def __init__(self, cursor, retry_attempts: int, retryable_exceptions: List[str]):
        self._cursor = cursor
        self._retry_attempts = retry_attempts
        self._retryable_exceptions = retryable_exceptions

    def execute(self, sql: str, bindings=None):
        attempt, success, exc = 0, False, None
        while not success and attempt < self._retry_attempts:
            try:
                if bindings is None:
                    self._cursor.execute(sql)
                else:
                    self._cursor.execute(sql, bindings)
                success = True
            except Exception as e:
                exception_name = type(e).__name__
                if exception_name in self._retryable_exceptions:
                    time.sleep(2**attempt)
                    exc = e
                    attempt += 1
                else:
                    print(f"Did not retry exception named '{exception_name}'")
                    raise e
        if not success:
            if exc:
                raise exc
            else:
                raise RuntimeError(
                    "execute call failed, but no exceptions raised- this should be impossible"
                )
        return self

    # forward along all non-execute() methods/attribute look-ups
    def __getattr__(self, name):
        return getattr(self._cursor, name)


class Environment(abc.ABC):
    """An Environment is an abstraction to describe *where* the code you execute in your dbt-duckdb project
    actually runs. This could be the local Python process that runs dbt (which is the default),
    a remote server (like a Buena Vista instance), or even a Jupyter notebook kernel.
    """

    def __init__(self, creds: DuckDBCredentials):
        self._creds = creds

        # Add any module paths to the Python path for the environment
        if creds.module_paths:
            for path in creds.module_paths:
                if path not in sys.path:
                    sys.path.append(path)

        major, minor, patch = [int(x) for x in duckdb.__version__.split("-")[0].split(".")]
        if major == 0 and (minor < 10 or (minor == 10 and patch == 0)):
            self._supports_comments = False
        else:
            self._supports_comments = True

    @property
    def creds(self) -> DuckDBCredentials:
        return self._creds

    @abc.abstractmethod
    def handle(self):
        pass

    @abc.abstractmethod
    def submit_python_job(self, handle, parsed_model: dict, compiled_code: str) -> AdapterResponse:
        pass

    @abc.abstractmethod
    def load_source(self, plugin_name: str, source_config: SourceConfig) -> str:
        pass

    @abc.abstractmethod
    def store_relation(self, plugin_name: str, target_config: TargetConfig) -> None:
        pass

    def get_binding_char(self) -> str:
        return "?"

    def supports_comments(self) -> bool:
        return self._supports_comments

    @classmethod
    @abc.abstractmethod
    def is_cancelable(cls) -> bool:
        pass

    @classmethod
    @abc.abstractmethod
    def cancel(cls, connection: Connection):
        pass

    @classmethod
    def initialize_db(
        cls, creds: DuckDBCredentials, plugins: Optional[Dict[str, BasePlugin]] = None
    ):
        config = creds.config_options or {}
        plugins = plugins or {}
        for plugin in plugins.values():
            plugin.update_connection_config(creds, config)

        if creds.retries:
            success, attempt, exc = False, 0, None
            while not success and attempt < creds.retries.connect_attempts:
                try:
                    conn = duckdb.connect(creds.path, read_only=False, config=config)
                    success = True
                except Exception as e:
                    exception_name = type(e).__name__
                    if exception_name in creds.retries.retryable_exceptions:
                        time.sleep(2**attempt)
                        exc = e
                        attempt += 1
                    else:
                        print(f"Did not retry exception named '{exception_name}'")
                        raise e
            if not success:
                if exc:
                    raise exc
                else:
                    raise RuntimeError(
                        "connect call failed, but no exceptions raised- this should be impossible"
                    )

        else:
            conn = duckdb.connect(creds.path, read_only=False, config=config)

        # install any extensions on the connection
        if creds.extensions is not None:
            for extension in creds.extensions:
                if isinstance(extension, str):
                    conn.install_extension(extension)
                    conn.load_extension(extension)
                elif isinstance(extension, dict):
                    try:
                        ext = Extension(**extension)
                    except Exception as e:
                        raise DbtRuntimeError(f"Failed to parse extension: {e}")
                    conn.execute(f"install {ext.name} from {ext.repo}")
                    conn.load_extension(ext.name)

        # Create/update secrets on the database
        for sql in creds.secrets_sql():
            conn.execute(sql)

        # Attach any fsspec filesystems on the database
        if creds.filesystems:
            import fsspec

            for spec in creds.filesystems:
                curr = spec.copy()
                fsimpl = curr.pop("fs")
                fs = fsspec.filesystem(fsimpl, **curr)
                conn.register_filesystem(fs)

        # let the plugins do any configuration on the
        # connection that they need to do
        if plugins:
            for plugin in plugins.values():
                plugin.configure_connection(conn)

        # attach any databases that we will be using
        if creds.attach:
            for attachment in creds.attach:
                conn.execute(attachment.to_sql())

        return conn

    @classmethod
    def initialize_cursor(
        cls,
        creds: DuckDBCredentials,
        cursor,
        plugins: Optional[Dict[str, BasePlugin]] = None,
        registered_df: dict = {},
    ):
        if creds.settings is not None:
            for key, value in creds.settings.items():
                # Okay to set these as strings because DuckDB will cast them
                # to the correct type
                cursor.execute(f"SET {key} = '{value}'")

        # update cursor if something is lost in the copy
        # of the parent connection
        if plugins:
            for plugin in plugins.values():
                plugin.configure_cursor(cursor)

        for df_name, df in registered_df.items():
            cursor.register(df_name, df)

        if creds.retries and creds.retries.query_attempts:
            cursor = RetryableCursor(
                cursor, creds.retries.query_attempts, creds.retries.retryable_exceptions
            )

        return cursor

    @classmethod
    def initialize_plugins(cls, creds: DuckDBCredentials) -> Dict[str, BasePlugin]:
        ret = {}
        base_config = creds.settings or {}
        for plugin_def in creds.plugins or []:
            config = base_config.copy()
            config.update(plugin_def.config or {})
            plugin = BasePlugin.create(
                plugin_def.module, config=config, alias=plugin_def.alias, credentials=creds
            )
            ret[plugin.name] = plugin
        return ret

    @classmethod
    def run_python_job(
        cls, con, load_df_function, identifier: str, compiled_code: str, creds: DuckDBCredentials
    ):
        mod_file = tempfile.NamedTemporaryFile(suffix=".py", delete=False)
        mod_file.write(compiled_code.lstrip().encode("utf-8"))
        mod_file.close()

        # Ensure that we have an event loop for async code to use since we may
        # be running inside of a thread that doesn't have one defined
        _ensure_event_loop()

        try:
            spec = importlib.util.spec_from_file_location(identifier, mod_file.name)
            if not spec:
                raise DbtRuntimeError(
                    "Failed to load python model as module: {}".format(identifier)
                )
            module = importlib.util.module_from_spec(spec)
            if spec.loader:
                spec.loader.exec_module(module)
            else:
                raise DbtRuntimeError(
                    "Python module spec is missing loader: {}".format(identifier)
                )

            # Do the actual work to run the code here
            dbt = module.dbtObj(load_df_function)
            df = module.model(dbt, con)
            if isinstance(df, duckdb.DuckDBPyRelation):
                # a duckdb relation might contain references to temporary tables
                # that cannot cross cursor boundaries
                module.materialize(df, con)
            else:
                # Create a separate read cursor to enable batched reads/writes
                cur = cls.initialize_cursor(creds, con.cursor())
                module.materialize(df, cur)
        except Exception as err:
            raise DbtRuntimeError(f"Python model failed:\n" f"{err}")
        finally:
            os.unlink(mod_file.name)


def create(creds: DuckDBCredentials) -> Environment:
    """Create an Environment based on the credentials passed in."""

    if creds.remote:
        from .buenavista import BVEnvironment

        return BVEnvironment(creds)
    elif creds.is_motherduck:
        from .motherduck import MotherDuckEnvironment

        return MotherDuckEnvironment(creds)
    else:
        from .local import LocalEnvironment

        return LocalEnvironment(creds)



import threading

from dbt_common.exceptions import DbtRuntimeError

from . import Environment
from .. import credentials
from .. import utils
from dbt.adapters.contracts.connection import AdapterResponse
from dbt.adapters.contracts.connection import Connection


class DuckDBCursorWrapper:
    def __init__(self, cursor):
        self._cursor = cursor

    # forward along all non-execute() methods/attribute look-ups
    def __getattr__(self, name):
        return getattr(self._cursor, name)

    def execute(self, sql, bindings=None):
        try:
            if bindings is None:
                return self._cursor.execute(sql)
            else:
                return self._cursor.execute(sql, bindings)
        except RuntimeError as e:
            raise DbtRuntimeError(str(e))


class DuckDBConnectionWrapper:
    def __init__(self, cursor, env):
        self._cursor = DuckDBCursorWrapper(cursor)
        self._env = env

    def close(self):
        self._cursor.close()
        self._env.notify_closed()

    def cursor(self):
        return self._cursor


class LocalEnvironment(Environment):
    def __init__(self, credentials: credentials.DuckDBCredentials):
        # Set the conn attribute to None so it always exists even if
        # DB initialization fails
        super().__init__(credentials)
        self.conn = None
        self._plugins = self.initialize_plugins(credentials)
        self.handle_count = 0
        self.lock = threading.RLock()
        self._keep_open = (
            credentials.keep_open or credentials.path == ":memory:" or credentials.is_motherduck
        )
        self._REGISTERED_DF: dict = {}

    def notify_closed(self):
        with self.lock:
            self.handle_count -= 1
            if self.handle_count == 0 and not self._keep_open:
                self.close()

    def is_cancelable(cls):
        return True

    @classmethod
    def cancel(cls, connection: Connection):
        connection.handle.cursor().interrupt()

    def handle(self):
        # Extensions/settings need to be configured per cursor
        with self.lock:
            if self.conn is None:
                self.conn = self.initialize_db(self.creds, self._plugins)
            self.handle_count += 1

        cursor = self.initialize_cursor(
            self.creds, self.conn.cursor(), self._plugins, self._REGISTERED_DF
        )
        return DuckDBConnectionWrapper(cursor, self)

    def submit_python_job(self, handle, parsed_model: dict, compiled_code: str) -> AdapterResponse:
        con = handle.cursor()

        def ldf(table_name):
            return con.query(f"select * from {table_name}")

        self.run_python_job(con, ldf, parsed_model["alias"], compiled_code, self.creds)
        return AdapterResponse(_message="OK")

    def load_source(self, plugin_name: str, source_config: utils.SourceConfig):
        if plugin_name not in self._plugins:
            raise Exception(
                f"Plugin {plugin_name} not found; known plugins are: "
                + ",".join(self._plugins.keys())
            )
        plugin = self._plugins[plugin_name]
        handle = self.handle()
        cursor = handle.cursor()

        if source_config.schema:
            cursor.execute(f"CREATE SCHEMA IF NOT EXISTS {source_config.schema}")

        save_mode = source_config.get("save_mode", "overwrite")
        if save_mode in ("ignore", "error_if_exists"):
            params = [source_config.schema, source_config.identifier]
            q = """SELECT COUNT(1)
                FROM system.information_schema.tables
                WHERE table_schema = ?
                AND table_name = ?
                """
            if source_config.database:
                q += "AND table_catalog = ?"
                params.append(source_config.database)
            if cursor.execute(q, params).fetchone()[0]:
                if save_mode == "error_if_exists":
                    raise Exception(f"Source {source_config.table_name()} already exists!")
                else:
                    # Nothing to do (we ignore the existing table)
                    return
        df = plugin.load(source_config)
        assert df is not None

        materialization = source_config.meta.get(
            "materialization", plugin.default_materialization()
        )
        source_table_name = source_config.table_name()
        df_name = source_table_name.replace(".", "_") + "_df"

        cursor.register(df_name, df)

        if materialization == "view":
            # save to df instance to register on each cursor creation
            self._REGISTERED_DF[df_name] = df

        cursor.execute(
            f"CREATE OR REPLACE {materialization} {source_table_name} AS SELECT * FROM {df_name}"
        )

        cursor.close()
        handle.close()

    def store_relation(self, plugin_name: str, target_config: utils.TargetConfig) -> None:
        if plugin_name not in self._plugins:
            if plugin_name.startswith("glue|"):
                from ..plugins import glue

                _, glue_db = plugin_name.split("|")
                config = (self.creds.settings or {}).copy()
                config["glue_database"] = glue_db
                self._plugins[plugin_name] = glue.Plugin(
                    name=plugin_name, plugin_config=config, credentials=self.creds
                )
            else:
                raise Exception(
                    f"Plugin {plugin_name} not found; known plugins are: "
                    + ",".join(self._plugins.keys())
                )
        plugin = self._plugins[plugin_name]
        plugin.store(target_config)

    def close(self):
        if self.conn:
            self.conn.close()
            self.conn = None

    def __del__(self):
        self.close()



from dbt.tests.adapter.simple_snapshot.test_snapshot import (
    BaseSnapshotCheck,
    BaseSimpleSnapshot,
)


class TestSimpleSnapshotDuckDB(BaseSimpleSnapshot):
    pass


class TestSnapshotCheckDuckDB(BaseSnapshotCheck):
    pass



import pytest
from dbt.tests.util import (
    check_relation_types,
    check_relations_equal,
    check_result_nodes_by_name,
    relation_from_name,
    run_dbt,
)

@pytest.mark.skip_profile("buenavista", "nightly", reason="Cannot install community extensions for nightly release")
class BaseCommunityExtensions:

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target):
        dbt_profile_target["extensions"] = [
            {"name": "quack", "repo": "community"},
        ]
        return dbt_profile_target

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "quack_model.sql": "select quack('world') as quack_world",
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "base",
        }

    def test_base(self, project):

        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 1

        # names exist in result nodes
        check_result_nodes_by_name(
            results,
            [
                "quack_model",
            ],
        )

        # check relation types
        expected = {
            "quack_model": "view",
        }
        check_relation_types(project.adapter, expected)

@pytest.mark.skip_profile("nightly", reason="Cannot install community extensions for nightly release")
@pytest.mark.skip_profile("buenavista")
class TestCommunityExtensions(BaseCommunityExtensions):
    pass



import pytest

from dbt.tests.util import (
    run_dbt,
)

my_date_spine_model = """
{{
    config(
        materialized = 'table',
    )
}}

with days as (

    {{
        dbt_utils.date_spine(
            'day',
            "'2024-01-01'::timestamp",
            dbt.dateadd('day', 1, "'2024-02-01'::timestamp"),
        )
    }}

),

final as (
    select cast(date_day as date) as date_day
    from days
)

select * from final
"""

class TestDateSpine:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "date_spine.sql": my_date_spine_model,
        }
    
    @pytest.fixture(scope="class")
    def packages(self):
        return {"packages": [{"package": "dbt-labs/dbt_utils", "version": "1.1.1"}]}

    def test_date_spine(self, project):

        # install dbt_utils
        run_dbt(["deps"])
        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 1



import pytest
from unittest.mock import patch

from duckdb.duckdb import IOException

from dbt.adapters.duckdb.credentials import DuckDBCredentials
from dbt.adapters.duckdb.credentials import Retries
from dbt.adapters.duckdb.environments import Environment

class TestConnectRetries:

    @pytest.fixture
    def creds(self):
        # Create a mock credentials object
        return DuckDBCredentials(
            path="foo.db",
            retries=Retries(connect_attempts=2, retryable_exceptions=["IOException", "ArithmeticError"])
        )

    @pytest.mark.parametrize("exception", [None, IOException, ArithmeticError, ValueError])
    def test_initialize_db(self, creds, exception):
        # Mocking the duckdb.connect method
        with patch('duckdb.connect') as mock_connect:
            if exception:
                mock_connect.side_effect = [exception, None]

            if exception == ValueError:
                with pytest.raises(ValueError) as excinfo:
                    Environment.initialize_db(creds)
            else:
                # Call the initialize_db method
                Environment.initialize_db(creds)
                if exception in {IOException, ArithmeticError}:
                    assert mock_connect.call_count == creds.retries.connect_attempts
                else:
                    mock_connect.assert_called_once_with(creds.path, read_only=False, config={})



import os
import pytest
from dbt.tests.adapter.basic.files import (
    base_table_sql,
    model_base,
    schema_base_yml,
    seeds_base_csv,
)
from dbt.tests.util import (
    run_dbt,
)

config_materialized_glue = """
  {{ config(materialized="external", glue_register=true, glue_database='db2') }}
"""
default_glue_sql = config_materialized_glue + model_base


@pytest.mark.skip
class TestGlueMaterializations:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "table_model.sql": base_table_sql,
            "table_default.sql": default_glue_sql,
            "schema.yml": schema_base_yml,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
        }

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target):
        dbt_profile_target["external_root"] = "s3://duckdbtest/glue_test"
        dbt_profile_target["extensions"] = ["httpfs"]
        dbt_profile_target["settings"] = {
            "s3_access_key_id": os.getenv("AWS_ACCESS_KEY_ID"),
            "s3_secret_access_key": os.getenv("AWS_SECRET_ACCESS_KEY"),
            "s3_region": "us-west-2",
        }
        return dbt_profile_target

    def test_base(self, project):
        # seed command
        results = run_dbt(["seed"])
        # seed result length
        assert len(results) == 1

        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 2



import pytest

from dbt.tests.adapter.constraints.test_constraints import (
    BaseTableConstraintsColumnsEqual,
    BaseViewConstraintsColumnsEqual,
    BaseIncrementalConstraintsColumnsEqual,
    BaseConstraintsRuntimeDdlEnforcement,
    BaseConstraintsRollback,
    BaseIncrementalConstraintsRuntimeDdlEnforcement,
    BaseIncrementalConstraintsRollback,
    BaseModelConstraintsRuntimeEnforcement,
)


class DuckDBColumnEqualSetup:
    @pytest.fixture
    def int_type(self):
        return "INT"

    @pytest.fixture
    def string_type(self):
        return "VARCHAR"

    @pytest.fixture
    def data_types(self, schema_int_type, int_type, string_type):
        # sql_column_value, schema_data_type, error_data_type
        return [
            ["1", schema_int_type, int_type],
            ["'1'", string_type, string_type],
            ["true", "bool", "BOOL"],
            ["'2013-11-03 00:00:00-07'::timestamp", "TIMESTAMP", "TIMESTAMP"],
            ["'2013-11-03 00:00:00-07'::timestamptz", "TIMESTAMPTZ", "TIMESTAMP WITH TIME ZONE"],
            ["ARRAY['a','b','c']", "VARCHAR[]", "VARCHAR[]"],
            ["ARRAY[1,2,3]", "INTEGER[]", "INTEGER[]"],
            ["'1'::numeric", "numeric", "DECIMAL"],
            [
                """'{"bar": "baz", "balance": 7.77, "active": false}'::json""",
                "json",
                "JSON",
            ],
        ]


class TestTableConstraintsColumnsEqual(
    DuckDBColumnEqualSetup, BaseTableConstraintsColumnsEqual
):
    pass


class TestViewConstraintsColumnsEqual(
    DuckDBColumnEqualSetup, BaseViewConstraintsColumnsEqual
):
    pass


class TestIncrementalConstraintsColumnsEqual(
    DuckDBColumnEqualSetup, BaseIncrementalConstraintsColumnsEqual
):
    pass


@pytest.mark.skip_profile("md")
class TestTableConstraintsRuntimeDdlEnforcement(
    DuckDBColumnEqualSetup, BaseConstraintsRuntimeDdlEnforcement
):
    pass


@pytest.mark.skip_profile("md", "buenavista")
class TestTableConstraintsRollback(DuckDBColumnEqualSetup, BaseConstraintsRollback):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NOT NULL constraint failed"]


@pytest.mark.skip_profile("md")
class TestIncrementalConstraintsRuntimeDdlEnforcement(
    DuckDBColumnEqualSetup, BaseIncrementalConstraintsRuntimeDdlEnforcement
):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NOT NULL constraint failed"]


@pytest.mark.skip_profile("md", "buenavista")
class TestIncrementalConstraintsRollback(
    DuckDBColumnEqualSetup, BaseIncrementalConstraintsRollback
):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NOT NULL constraint failed"]


@pytest.mark.skip_profile("md")
class TestModelConstraintsRuntimeEnforcement(
    DuckDBColumnEqualSetup, BaseModelConstraintsRuntimeEnforcement
):
    @pytest.fixture(scope="class")
    def expected_error_messages(self):
        return ["NOT NULL constraint failed"]



import os
from uuid import uuid4
import boto3
import pytest
from dbt.tests.adapter.basic.files import (
    base_table_sql,
    model_base,
    schema_base_yml,
    seeds_base_csv,
)
from dbt.tests.util import (
    check_relation_types,
    check_relations_equal,
    check_result_nodes_by_name,
    relation_from_name,
    run_dbt,
)

config_materialized_default = """
  {{ config(materialized="external") }}
"""

config_materialized_csv = """
  {{ config(materialized="external", format="csv") }}
"""

config_materialized_parquet_location = """
  {{ config(materialized="external", location="{{ adapter.external_root() }}/test.parquet") }}
"""

config_materialized_csv_location_delim = """
  {{ config(materialized="external", location="{{ adapter.external_root() }}/test_delim.csv", delimiter="|") }}
"""

config_json = """
  {{ config(materialized="external", location="{{ adapter.external_root() }}/test.json") }}
"""

default_external_sql = config_materialized_default + model_base
csv_external_sql = config_materialized_csv + model_base
parquet_table_location_sql = config_materialized_parquet_location + model_base
csv_location_delim_sql = config_materialized_csv_location_delim + model_base
json_sql = config_json + model_base
make_empty = " where id > 10"

DEST_LOCAL = "local"
DEST_S3 = "s3"
BUCKET_NAME = "md-ecosystem-public"

class BaseExternalMaterializations:
    @pytest.fixture(scope="class")
    def models(self, empty):
        return {
            "table_model.sql": base_table_sql + (make_empty if empty else ""),
            "table_default.sql": default_external_sql + (make_empty if empty else ""),
            "table_csv.sql": csv_external_sql + (make_empty if empty else ""),
            "table_parquet_location.sql": parquet_table_location_sql + (make_empty if empty else ""),
            "table_csv_location_delim.sql": csv_location_delim_sql + (make_empty if empty else ""),
            "table_json.sql": json_sql + (make_empty if empty else ""),
            "schema.yml": schema_base_yml,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "base.csv": seeds_base_csv,
        }

    @pytest.fixture(scope="class")
    def dest(self):
        return DEST_LOCAL

    @pytest.fixture(scope="class")
    def empty(self):
        return False

    @pytest.fixture(scope="class")
    def extroot(self, tmp_path_factory, dest):
        if dest == DEST_LOCAL:
            extroot = str(tmp_path_factory.getbasetemp() / "external")
            if not os.path.exists(extroot):
                os.mkdir(extroot)

        elif dest == DEST_S3:
            s3_path = os.path.join("dbt-duckdb", "test_external", f"pytest-{uuid4()}")
            extroot = os.path.join("s3://", BUCKET_NAME, s3_path)

        yield extroot

        if dest == DEST_S3:
            session = boto3.Session(
                aws_access_key_id=os.getenv("S3_MD_ORG_KEY"),
                aws_secret_access_key=os.getenv("S3_MD_ORG_SECRET"),
            )
            s3 = session.resource('s3')
            bucket = s3.Bucket('md-ecosystem-public')
            bucket.objects.filter(Prefix=s3_path).delete()

    @pytest.fixture(scope="class")
    def dbt_profile_target(self, profile_type, dbt_profile_target, extroot):
        dbt_profile_target["external_root"] = extroot
        dbt_profile_target["secrets"] = [
            {
                "type": DEST_S3,
                "region": os.getenv("S3_MD_ORG_REGION"),
                "key_id": os.getenv("S3_MD_ORG_KEY"),
                "secret": os.getenv("S3_MD_ORG_SECRET"),
            }
        ]
        if profile_type == "md":
            dbt_profile_target["secrets"][0]["persistent"] = True
        return dbt_profile_target

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "name": "base",
        }

    def test_base(self, project, empty):

        # seed command
        results = run_dbt(["seed"])
        # seed result length
        assert len(results) == 1

        # run command
        results = run_dbt()
        # run result length
        assert len(results) == 6

        # names exist in result nodes
        check_result_nodes_by_name(
            results,
            [
                "table_model",
                "table_default",
                "table_csv",
                "table_parquet_location",
                "table_csv_location_delim",
                "table_json",
            ],
        )

        # check relation types
        expected = {
            "base": "table",
            "table_model": "table",
            "table_default": "view",
            "table_parquet_location": "view",
            "table_csv": "view",
            "table_csv_location_delim": "view",
            "table_json": "view",
        }
        if empty:
            # if simulating an empty table, result will be empty
            expected.pop("base")
        check_relation_types(project.adapter, expected)

        # base table rowcount
        relation = relation_from_name(project.adapter, "base")
        result = project.run_sql(f"select count(*) as num_rows from {relation}", fetch="one")
        assert result[0] == 10

        # relations_equal
        models = [
            "base",
            "table_default",
            "table_parquet_location",
            "table_model",
            "table_csv",
            "table_csv_location_delim",
            "table_json",
        ]
        if empty:
            # if simulating an empty table, result will be empty, so don't compare to base table
            models.pop(0)
        check_relations_equal(
            project.adapter,
            models,
        )

        # check relations in catalog
        catalog = run_dbt(["docs", "generate"])
        assert len(catalog.nodes) == 7
        assert len(catalog.sources) == 1


@pytest.mark.skip_profile("buenavista")
class TestExternalMaterializations(BaseExternalMaterializations):
    pass


@pytest.mark.skip_profile("buenavista")
class TestExternalMaterializationsLocalEmpty(BaseExternalMaterializations):
    @pytest.fixture(scope="class")
    def dest(self):
        return DEST_LOCAL
    
    @pytest.fixture(scope="class")
    def empty(self):
        return True


@pytest.mark.with_s3_creds
@pytest.mark.skip_profile("buenavista")
class TestExternalMaterializationsS3(BaseExternalMaterializations):
    @pytest.fixture(scope="class")
    def dest(self):
        return DEST_S3
    
    @pytest.fixture(scope="class")
    def empty(self):
        return False


@pytest.mark.with_s3_creds
@pytest.mark.skip_profile("buenavista")
class TestExternalMaterializationsS3Empty(BaseExternalMaterializations):
    @pytest.fixture(scope="class")
    def dest(self):
        return DEST_S3
    
    @pytest.fixture(scope="class")
    def empty(self):
        return True



import pytest
from dbt.tests.util import run_dbt
from dbt.adapters.duckdb.connections import DuckDBConnectionManager

models_file_model_sql = """
{{ config(materialized='table') }}
select *
from read_csv_auto('github://data/team_ratings.csv')
WHERE conf = 'West'
"""


@pytest.mark.skip_profile("buenavista", "md")
class TestFilesystems:
    @pytest.fixture(scope="class")
    def dbt_profile_target(self, dbt_profile_target):
        return {
            "type": "duckdb",
            "path": dbt_profile_target.get("path", ":memory:"),
            "filesystems": [
                {"fs": "github", "org": "jwills", "repo": "nba_monte_carlo"}
            ],
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "file_model.sql": models_file_model_sql,
        }

    def test_filesystems(self, project):
        DuckDBConnectionManager.close_all_connections()
        results = run_dbt()
        assert len(results) == 1



import atexit
import threading
from contextlib import contextmanager
from multiprocessing.context import SpawnContext
from typing import Optional
from typing import Set
from typing import Tuple
from typing import TYPE_CHECKING

import dbt.exceptions
from . import environments
from dbt.adapters.contracts.connection import AdapterRequiredConfig
from dbt.adapters.contracts.connection import AdapterResponse
from dbt.adapters.contracts.connection import Connection
from dbt.adapters.contracts.connection import ConnectionState
from dbt.adapters.events.logging import AdapterLogger
from dbt.adapters.sql import SQLConnectionManager

logger = AdapterLogger("DuckDB")

if TYPE_CHECKING:
    import agate


class DuckDBConnectionManager(SQLConnectionManager):
    TYPE = "duckdb"
    _LOCK = threading.RLock()
    _ENV = None
    _LOGGED_MESSAGES: Set[str] = set()

    def __init__(self, config: AdapterRequiredConfig, mp_context: SpawnContext) -> None:
        super().__init__(config, mp_context)
        self.disable_transactions = config.credentials.disable_transactions  # type: ignore

    @classmethod
    def env(cls) -> environments.Environment:
        with cls._LOCK:
            if not cls._ENV:
                raise Exception("DuckDBConnectionManager environment requested before creation!")
            return cls._ENV

    @classmethod
    def open(cls, connection: Connection) -> Connection:
        if connection.state == ConnectionState.OPEN:
            logger.debug("Connection is already open, skipping open.")
            return connection

        credentials = cls.get_credentials(connection.credentials)
        with cls._LOCK:
            try:
                if not cls._ENV or cls._ENV.creds != credentials:
                    cls._ENV = environments.create(credentials)
                connection.handle = cls._ENV.handle()
                connection.state = ConnectionState.OPEN

            except RuntimeError as e:
                logger.debug("Got an error when attempting to connect to DuckDB: '{}'".format(e))
                connection.handle = None
                connection.state = ConnectionState.FAIL
                raise dbt.adapters.exceptions.FailedToConnectError(str(e))

            return connection

    @classmethod
    def close(cls, connection: Connection) -> Connection:
        # if the connection is in closed or init, there's nothing to do
        if connection.state in {ConnectionState.CLOSED, ConnectionState.INIT}:
            return connection

        connection = super(SQLConnectionManager, cls).close(connection)
        return connection

    @classmethod
    def warn_once(cls, msg: str):
        """Post a warning message once per dbt execution."""
        with cls._LOCK:
            if msg in cls._LOGGED_MESSAGES:
                return
            cls._LOGGED_MESSAGES.add(msg)
            logger.warning(msg)

    def cancel(self, connection: Connection):
        if self._ENV is not None:
            logger.debug(
                "cancelling query on connection {}. Details: {}".format(
                    connection.name, connection
                )
            )
            self._ENV.cancel(connection)
            logger.debug("query cancelled on connection {}".format(connection.name))

    @contextmanager
    def exception_handler(self, sql: str, connection_name="master"):
        try:
            yield
        except dbt.exceptions.DbtRuntimeError:
            raise
        except RuntimeError as e:
            logger.debug("duckdb error: {}".format(str(e)))
        except Exception as exc:
            logger.debug("Error running SQL: {}".format(sql))
            logger.debug("Rolling back transaction.")
            raise dbt.exceptions.DbtRuntimeError(str(exc)) from exc

    @classmethod
    def get_credentials(cls, credentials):
        return credentials

    @classmethod
    def get_response(cls, cursor) -> AdapterResponse:
        # https://github.com/dbt-labs/dbt-spark/issues/142
        message = "OK"
        return AdapterResponse(_message=message)

    @classmethod
    def close_all_connections(cls):
        with cls._LOCK:
            if cls._ENV is not None:
                cls._ENV = None

    def execute(
        self,
        sql: str,
        auto_begin: bool = False,
        fetch: bool = False,
        limit: Optional[int] = None,
    ) -> Tuple[AdapterResponse, "agate.Table"]:
        if self.disable_transactions:
            auto_begin = False
        return super().execute(sql, auto_begin, fetch, limit)


atexit.register(DuckDBConnectionManager.close_all_connections)



import pytest
import sqlite3
from pathlib import Path
from dbt.tests.util import (
    run_dbt,
)

model_sql = """
    {{ config(materialized='incremental', database='satest') }}
    select * from satest.tt1
"""


class TestSQLitePlugin:

    @pytest.fixture(scope="class")
    def sqlite_test_db(self):
        path = '/tmp/satest.db'
        Path(path).unlink(missing_ok=True)
        db = sqlite3.connect(path)
        cursor = db.cursor()
        cursor.execute("CREATE TABLE tt1 (id int, name text)")
        cursor.execute("INSERT INTO tt1 VALUES (1, 'John Doe')")
        cursor.execute("INSERT INTO tt1 VALUES (2, 'Jane Smith')")
        cursor.execute("CREATE TABLE test_table2 (a int, b int, c int)")
        cursor.execute("INSERT INTO test_table2 VALUES (1, 2, 3), (4, 5, 6)")
        cursor.close()
        db.commit()
        db.close()

        yield path

    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target, sqlite_test_db):
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target.get("path", ":memory:"),
                        "attach": [
                           {'path': sqlite_test_db}
                        ]
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self, test_data_path):
        return {
            "read_write.sql": model_sql,

        }

    def test_sqlite_plugin(self, project):
        results = run_dbt()
        assert len(results) == 1

        res = project.run_sql("SELECT COUNT(1) FROM satest.read_write", fetch="one")
        assert res[0] == 2





import os
import pathlib
from threading import Lock
from typing import Any
from typing import Dict

import pandas as pd
from pandas.io.formats import excel

from . import BasePlugin
from . import pd_utils
from ..utils import SourceConfig
from ..utils import TargetConfig
from dbt.adapters.events.logging import AdapterLogger

logger = AdapterLogger("DuckDB")


class Plugin(BasePlugin):
    def initialize(self, plugin_config: Dict[str, Any]):
        self._config = plugin_config

        if "output" in plugin_config:
            self._excel_writer_create_lock = Lock()
            assert isinstance(plugin_config["output"], dict)
            assert "file" in plugin_config["output"]

        # Pass s3 settings to plugin environment
        if "s3_access_key_id" in plugin_config:
            os.environ["AWS_ACCESS_KEY_ID"] = plugin_config["s3_access_key_id"]
        if "s3_secret_access_key" in plugin_config:
            os.environ["AWS_SECRET_ACCESS_KEY"] = plugin_config["s3_secret_access_key"]
        if "s3_region" in plugin_config:
            os.environ["AWS_DEFAULT_REGION"] = plugin_config["s3_region"]

    def load(self, source_config: SourceConfig):
        ext_location = source_config["external_location"]
        ext_location = ext_location.format(**source_config.as_dict())
        if "s3" in ext_location:
            # Possible to add some treatment in the future
            source_location = ext_location
        else:
            source_location = pathlib.Path(ext_location.strip("'"))
        sheet_name = source_config.get("sheet_name", 0)
        return pd.read_excel(source_location, sheet_name=sheet_name)

    def store(self, target_config: TargetConfig):
        plugin_output_config = self._config["output"]

        # Create the writer on the first instance of the call to store.
        # Instead if we instantiated the writer in the constructor
        # with mode = 'w', this would result in an existing file getting
        # overwritten. This can happen if dbt test is executed for example.
        if not hasattr(self, "_excel_writer"):
            with self._excel_writer_create_lock:
                if not hasattr(self, "_excel_writer"):
                    self._excel_writer = pd.ExcelWriter(
                        plugin_output_config["file"],
                        mode=plugin_output_config.get("mode", "w"),
                        engine=plugin_output_config.get("engine", "xlsxwriter"),
                        engine_kwargs=plugin_output_config.get("engine_kwargs", {}),
                        date_format=plugin_output_config.get("date_format"),
                        datetime_format=plugin_output_config.get("datetime_format"),
                    )
                    if not plugin_output_config.get("header_styling", True):
                        excel.ExcelFormatter.header_style = None

        target_output_config = {
            **plugin_output_config,
            **target_config.config.get("overrides", {}),
        }

        if "sheet_name" not in target_output_config:
            # Excel sheet name is limited to 31 characters
            sheet_name = (target_config.relation.identifier or "Sheet1")[0:31]
            target_output_config["sheet_name"] = sheet_name

        df = pd_utils.target_to_df(target_config)
        if target_output_config.get("skip_empty_sheet", False) and df.shape[0] == 0:
            return
        try:
            df.to_excel(
                self._excel_writer,
                sheet_name=target_output_config["sheet_name"],
                na_rep=target_output_config.get("na_rep", ""),
                float_format=target_output_config.get("float_format", None),
                header=target_output_config.get("header", True),
                index=target_output_config.get("index", True),
                merge_cells=target_output_config.get("merge_cells", True),
                inf_rep=target_output_config.get("inf_rep", "inf"),
            )
            if not target_output_config.get("lazy_close", True):
                self._excel_writer.close()
                del self._excel_writer
        except ValueError as ve:
            # Catches errors resembling the below & logs an appropriate message
            # ValueError('This sheet is too large! Your sheet size is: 1100000, 1 Max sheet size is: 1048576, 16384')
            if (
                str(ve).startswith("This sheet is too large")
                and target_output_config["ignore_sheet_too_large"]
            ):
                pd.DataFrame(
                    [{"Error": target_output_config.get("ignore_sheet_too_large_error", str(ve))}]
                ).to_excel(
                    self._excel_writer, sheet_name=target_output_config["sheet_name"], index=False
                )
            else:
                raise ve

    def __del__(self):
        if hasattr(self, "_excel_writer"):
            logger.info(f"Closing {self._config['output']['file']}")
            self._excel_writer.close()



import pytest
from pathlib import Path
import pandas as pd
import tempfile

from dbt.tests.util import (
    run_dbt,
)
from deltalake.writer import write_deltalake

delta_schema_yml = """
version: 2
sources:
  - name: delta_source
    meta:
      plugin: delta
    tables:
      - name: table_1
        description: "An delta table"
        meta:
          delta_table_path: "{test_delta_path1}"

  - name: delta_source_test
    schema: test
    meta:
      plugin: delta
    tables:
      - name: table_2
        description: "An delta table"
        meta:
          delta_table_path: "{test_delta_path2}"
          as_of_version: 0
"""


delta1_sql = """
    {{ config(materialized='table') }}
    select * from {{ source('delta_source', 'table_1') }}
"""
delta2_sql = """
    {{ config(materialized='table') }}
    select * from {{ source('delta_source', 'table_1') }} limit 1
"""
delta3_sql = """
    {{ config(materialized='table') }}
    select * as a from {{ source('delta_source_test', 'table_2') }} WHERE y = 'd'
"""

delta3_sql_expected = """
    select 1 as x, 'a' as y
"""


@pytest.mark.skip_profile("buenavista", "md")
class TestPlugins:
    @pytest.fixture(scope="class")
    def delta_test_table1(self):
        td = tempfile.TemporaryDirectory() 
        path = Path(td.name)
        table_path = path / "test_delta_table1"

        df = pd.DataFrame({"x": [1, 2, 3]})
        write_deltalake(table_path, df, mode="overwrite")

        yield table_path

        td.cleanup()

    @pytest.fixture(scope="class")
    def delta_test_table2(self):
        td = tempfile.TemporaryDirectory() 
        path = Path(td.name)
        table_path = path / "test_delta_table2"

        df = pd.DataFrame({
            "x": [1],
            "y": ["a"]                   
        })
        write_deltalake(table_path, df, mode="overwrite")

        df = pd.DataFrame({
            "x": [1, 2],
            "y": ["a","b"]                   
        })
        write_deltalake(table_path, df, mode="overwrite")

        yield table_path

        td.cleanup()

    @pytest.fixture(scope="class")
    def profiles_config_update(self, dbt_profile_target):
        plugins = [{"module": "delta"}]
        return {
            "test": {
                "outputs": {
                    "dev": {
                        "type": "duckdb",
                        "path": dbt_profile_target.get("path", ":memory:"),
                        "plugins": plugins,
                    }
                },
                "target": "dev",
            }
        }

    @pytest.fixture(scope="class")
    def models(self, delta_test_table1,delta_test_table2):
        return {
            "source_schema.yml": delta_schema_yml.format(
                test_delta_path1=delta_test_table1,
                test_delta_path2=delta_test_table2
            ),
            "delta_table1.sql": delta1_sql,
            "delta_table2.sql": delta2_sql,
            "delta_table3.sql": delta3_sql,
            "delta_table3_expected.sql": delta3_sql_expected,
        }

    def test_plugins(self, project):
        results = run_dbt()
        assert len(results) == 4

        # check_relations_equal(
        #     project.adapter,
        #     [
        #         "delta_table3",
        #         "delta_table3_expected",
        #     ],
        # )
        # res = project.run_sql("SELECT count(1) FROM 'delta_table3'", fetch="one")
        # assert res[0] == 2
